from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer, \
    kcwi_fits_reader, strip_fname

import time
import os
import math
import pickle
import numpy as np
from skimage import transform as tf
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.nddata import CCDData
from kcwidrp.core.bokeh_plotting import bokeh_plot
from bokeh.plotting import figure
from multiprocessing import Pool


def make_cube_helper(argument):
    """
    Warp each slice.
    
    Helper program for threaded warping of all slices.

    Args:
        argument (dict): Dictionary of params for warping an individual slice.

    :returns:
        - int: Slice number being warped
        - ndimage: warped intensity image
        - ndimage: warped uncertainty image
        - ndimage: warped mask image
        - ndimage: warped flag image
        - ndimage: warped image without sky subtraction
        - ndimage: warped nod-and-shuffle object image
        - ndimage: warped nod-and-shuffle sky image
        - ndimage: warped delta-wavelength image

    """
    logger = argument['logger']
    logger.info("Transforming image slice %d" % (argument['slice_number']+1))
    # input params
    slice_number = argument['slice_number']
    xsize = argument['xsize']
    ysize = argument['ysize']
    tform = argument['geom']['tform'][slice_number]
    xl0 = argument['geom']['xl0'][slice_number]
    xl1 = argument['geom']['xl1'][slice_number]
    # slice data
    slice_img = argument['img'][:, xl0:xl1]
    slice_unc = argument['std'][:, xl0:xl1]
    slice_msk = argument['msk'][:, xl0:xl1]
    slice_flg = argument['flg'][:, xl0:xl1]
    if 'nsk' in argument:
        slice_nsk = argument['nsk'][:, xl0:xl1]
    else:
        slice_nsk = None
    if 'obj' in argument:
        slice_obj = argument['obj'][:, xl0:xl1]
    else:
        slice_obj = None
    if 'sky' in argument:
        slice_sky = argument['sky'][:, xl0:xl1]
    else:
        slice_sky = None
    if 'del' in argument:
        slice_del = argument['del'][:, xl0:xl1]
    else:
        slice_del = None
    # do the warping
    warped = tf.warp(slice_img, tform, order=3,
                     output_shape=(ysize, xsize))
    uarped = tf.warp(slice_unc, tform, order=3,
                     output_shape=(ysize, xsize))
    marped = tf.warp(slice_msk, tform, order=3,
                     output_shape=(ysize, xsize))
    farped = tf.warp(slice_flg, tform, order=3,
                     output_shape=(ysize, xsize), preserve_range=True)

    if slice_nsk is not None:
        karped = tf.warp(slice_nsk, tform, order=3,
                         output_shape=(ysize, xsize))
    else:
        karped = None

    if slice_obj is not None:
        oarped = tf.warp(slice_obj, tform, order=3,
                         output_shape=(ysize, xsize))
    else:
        oarped = None

    if slice_sky is not None:
        sarped = tf.warp(slice_sky, tform, order=3,
                         output_shape=(ysize, xsize))
    else:
        sarped = None

    if slice_del is not None:
        darped = tf.warp(slice_del, tform, order=3,
                         output_shape=(ysize, xsize), preserve_range=True)
    else:
        darped = None

    return argument['slice_number'], warped, uarped, marped, farped, karped, \
        oarped, sarped, darped


class MakeCube(BasePrimitive):
    """
    Transform 2D images to 3D data cubes.

    Uses thread pool to warp each slice in parallel and then
    assemble a 3D image cube from the slices. Rotates RED channel
    data by 180 degress to align with BLUE channel data.

    Creates WCS keywords that reflect the new geometry of the
    cube based on position keywords from the telescope.

    Outputs FITS 3d cube images:

        * icube - main cube with primary, uncertainty, mask, flag, and noskysub extensions
        * ocube - for nod-and-shuffle, the object frame as a cube
        * scube - for nod-and-shuffle, the sky frame as a cube
        * dcube - the delta-wavelength cube

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _perform(self):
        self.logger.info("Creating data cube")

        log_string = MakeCube.__module__

        # Are we interactive?
        if self.config.instrument.plot_level >= 3:
            do_inter = True
        else:
            do_inter = False
        self.logger.info("Generating data cube")
        # Find and read geometry transformation
        tab = self.context.proctab.search_proctab(
                frame=self.action.args.ccddata, target_type='MARC',
                nearest=True)
        if not len(tab):
            self.logger.error("No reference geometry, cannot make cube!")
            self.action.args.ccddata.header['GEOMCOR'] = (False,
                                                          'Geometry corrected?')
            self.logger.info(log_string)
            return self.action.args

        self.logger.info("%d arc frames found" % len(tab))
        ofn = strip_fname(tab['filename'][0]) + "_geom.pkl"
        geom_file = os.path.join(self.config.instrument.cwd,
                                 self.config.instrument.output_directory,
                                 ofn)
        if os.path.exists(geom_file):
            self.logger.info("Reading %s" % geom_file)
            with open(geom_file, 'rb') as ifile:
                geom = pickle.load(ifile)
            # Slice size
            xsize = geom['xsize']
            ysize = geom['ysize']
            out_cube = np.zeros((ysize, xsize, 24), dtype=np.float64)
            out_uube = np.zeros((ysize, xsize, 24), dtype=np.float64)
            out_mube = np.zeros((ysize, xsize, 24), dtype=np.uint8)
            out_fube = np.zeros((ysize, xsize, 24), dtype=np.uint8)
            out_oube = np.zeros((ysize, xsize, 24), dtype=np.float64)
            out_sube = np.zeros((ysize, xsize, 24), dtype=np.float64)
            out_dube = np.zeros((ysize, xsize, 24), dtype=np.float64)
            # Store original data
            data_img = self.action.args.ccddata.data
            data_std = self.action.args.ccddata.uncertainty.array
            data_msk = self.action.args.ccddata.mask
            data_flg = self.action.args.ccddata.flags
            # check for noskysub property
            if self.action.args.ccddata.noskysub is not None:
                out_kube = np.zeros((ysize, xsize, 24), dtype=np.float64)
                data_nsk = self.action.args.ccddata.noskysub
            else:
                data_nsk = None
                out_kube = None
            # check for obj, sky images
            obj = None
            data_obj = None
            sky = None
            data_sky = None
            if self.action.args.nasmask and self.action.args.numopen > 1:
                ofn = self.action.args.name

                objfn = strip_fname(ofn) + '_objf.fits'
                full_path = os.path.join(
                    self.config.instrument.cwd,
                    self.config.instrument.output_directory, objfn)
                if os.path.exists(full_path):
                    obj = kcwi_fits_reader(full_path)[0]
                    data_obj = obj.data

                skyfn = strip_fname(ofn) + '_skyf.fits'
                full_path = os.path.join(
                    self.config.instrument.cwd,
                    self.config.instrument.output_directory, skyfn)
                if os.path.exists(full_path):
                    sky = kcwi_fits_reader(full_path)[0]
                    data_sky = sky.data
            # check for delta wavelength map
            dew = None
            data_dew = None
            if 'ARCLAMP' in self.action.args.imtype:
                ofn = self.action.args.name
                dewfn = strip_fname(ofn) + '_delmap.fits'
                full_path = os.path.join(
                    self.config.instrument.cwd,
                    self.config.instrument.output_directory, dewfn)
                if os.path.exists(full_path):
                    dew = kcwi_fits_reader(full_path)[0]
                    data_dew = dew.data
            # Loop over 24 slices
            my_arguments = []
            for isl in range(0, 24):
                arguments = {
                    'slice_number': isl,
                    'geom': geom,
                    'img': data_img,
                    'std': data_std,
                    'msk': data_msk,
                    'flg': data_flg,
                    'xsize': xsize,
                    'ysize': ysize,
                    'logger': self.logger
                }
                if data_nsk is not None:
                    arguments['nsk'] = data_nsk
                if obj is not None:
                    arguments['obj'] = data_obj
                if sky is not None:
                    arguments['sky'] = data_sky
                if dew is not None:
                    arguments['del'] = data_dew
                my_arguments.append(arguments)

            p = Pool()
            results = p.map(make_cube_helper, list(my_arguments))
            p.close()

            self.logger.info("Building cube")
            for partial_cube in results:
                slice_number = partial_cube[0]
                out_cube[:, :, slice_number] = partial_cube[1]
                out_uube[:, :, slice_number] = partial_cube[2]
                out_mube[:, :, slice_number] = partial_cube[3]
                out_fube[:, :, slice_number] = partial_cube[4]
                if data_nsk is not None:
                    out_kube[:, :, slice_number] = partial_cube[5]
                if obj is not None:
                    out_oube[:, :, slice_number] = partial_cube[6]
                if sky is not None:
                    out_sube[:, :, slice_number] = partial_cube[7]
                if dew is not None:
                    out_dube[:, :, slice_number] = partial_cube[8]

            if self.config.instrument.plot_level >= 3:
                for isl in range(0, 24):
                    warped = out_cube[:, :, isl]
                    ptitle = self.action.args.plotlabel + \
                        "WARPED Slice %d" % isl
                    p = figure(tooltips=[("x", "$x"), ("y", "$y"),
                                         ("value", "@image")],
                               title=ptitle,
                               x_axis_label="X (px)", y_axis_label="Y (px)",
                               plot_width=self.config.instrument.plot_width,
                               plot_height=self.config.instrument.plot_height)
                    p.x_range.range_padding = p.y_range.range_padding = 0
                    p.image([warped], x=0, y=0, dw=xsize, dh=ysize,
                            palette="Spectral11", level="image")
                    bokeh_plot(p, self.context.bokeh_session)
                    if do_inter:
                        q = input("Next? <cr>, q to quit: ")
                        if 'Q' in q.upper():
                            do_inter = False
                    else:
                        time.sleep(self.config.instrument.plot_pause)

            # Rotate RED data by 180 to align with Blue
            if 'RED' in self.action.args.ccddata.header['CAMERA'].upper():
                out_cube = np.rot90(out_cube, 2, (1, 2))
                out_uube = np.rot90(out_uube, 2, (1, 2))
                out_mube = np.rot90(out_mube, 2, (1, 2))
                out_fube = np.rot90(out_fube, 2, (1, 2))
                if data_nsk is not None:
                    out_kube = np.rot90(out_kube, 2, (1, 2))
                if obj is not None:
                    out_oube = np.rot90(out_oube, 2, (1, 2))
                if sky is not None:
                    out_sube = np.rot90(out_sube, 2, (1, 2))
                if dew is not None:
                    out_dube = np.rot90(out_dube, 2, (1, 2))

            # Calculate some WCS parameters
            # Get object pointing
            try:
                if self.action.args.nasmask:
                    rastr = self.action.args.ccddata.header['RABASE']
                    decstr = self.action.args.ccddata.header['DECBASE']
                else:
                    rastr = self.action.args.ccddata.header['RA']
                    decstr = self.action.args.ccddata.header['DEC']
            except KeyError:
                try:
                    rastr = self.action.args.ccddata.header['TARGRA']
                    decstr = self.action.args.ccddata.header['TARGDEC']
                except KeyError:
                    rastr = ''
                    decstr = ''
            if len(rastr) > 0 and len(decstr) > 0:
                try:
                    coord = SkyCoord(rastr, decstr, unit=(u.hourangle, u.deg))
                except ValueError:
                    coord = None
            else:
                coord = None
            # Get rotator position
            if 'ROTPOSN' in self.action.args.ccddata.header:
                rpos = self.action.args.ccddata.header['ROTPOSN']
            else:
                rpos = 0.
            if 'ROTREFAN' in self.action.args.ccddata.header:
                rref = self.action.args.ccddata.header['ROTREFAN']
            else:
                rref = 0.
            skypa = rpos + rref
            crota = math.radians(-(skypa + self.config.instrument.ROTOFF))
            cdelt1 = -geom['slscl']
            cdelt2 = geom['pxscl']
            if coord is None:
                ra = 0.
                dec = 0.
                crota = 1
            else:
                ra = coord.ra.degree
                dec = coord.dec.degree
            cd11 = cdelt1 * math.cos(crota)
            cd12 = abs(cdelt2) * np.sign(cdelt1) * math.sin(crota)
            cd21 = -abs(cdelt1) * np.sign(cdelt2) * math.sin(crota)
            cd22 = cdelt2 * math.cos(crota)
            crpix1 = 12.
            crpix2 = xsize / 2.
            crpix3 = 1.
            porg = self.action.args.ccddata.header['PONAME']
            ifunum = self.action.args.ifunum
            if 'IFU' in porg:
                if ifunum == 1:
                    off1 = 1.0
                    off2 = 4.0
                elif ifunum == 2:
                    off1 = 1.0
                    off2 = 5.0
                elif ifunum == 3:
                    off1 = 0.05
                    off2 = 5.6
                else:
                    self.logger.warning("Unknown IFU number: %d" % ifunum)
                    off1 = 0.
                    off2 = 0.
                off1 = off1 / float(self.action.args.xbinsize)
                off2 = off2 / float(self.action.args.ybinsize)
                crpix1 += off1
                crpix2 += off2
            # Update header
            # Geometry corrected?
            self.action.args.ccddata.header['GEOMCOR'] = (
                True, 'Geometry corrected?')
            #
            # Spatial geometry
            self.action.args.ccddata.header['BARSEP'] = (
                geom['barsep'], 'separation of bars (binned pix)')
            self.action.args.ccddata.header['BAR0'] = (
                geom['bar0'], 'first bar pixel position')
            # Wavelength ranges
            if self.action.args.nasmask:
                self.action.args.ccddata.header['WAVALL0'] = (
                    geom['wavensall0'], 'Low inclusive wavelength')
                self.action.args.ccddata.header['WAVALL1'] = (
                    geom['wavensall1'], 'High inclusive wavelength')
                self.action.args.ccddata.header['WAVGOOD0'] = (
                    geom['wavensgood0'], 'Low good wavelength')
                self.action.args.ccddata.header['WAVGOOD1'] = (
                    geom['wavensgood1'], 'High good wavelength')
                self.action.args.ccddata.header['WAVMID'] = (
                    geom['wavensmid'], 'middle wavelength')
            else:
                self.action.args.ccddata.header['WAVALL0'] = (
                    geom['waveall0'], 'Low inclusive wavelength')
                self.action.args.ccddata.header['WAVALL1'] = (
                    geom['waveall1'], 'High inclusive wavelength')
                self.action.args.ccddata.header['WAVGOOD0'] = (
                    geom['wavegood0'], 'Low good wavelength')
                self.action.args.ccddata.header['WAVGOOD1'] = (
                    geom['wavegood1'], 'High good wavelength')
                self.action.args.ccddata.header['WAVMID'] = (
                    geom['wavemid'], 'middle wavelength')
            # Dichroic fraction
            try:
                dichroic_fraction = geom['dich_frac']
            except AttributeError:
                dichroic_fraction = 1.
            self.action.args.ccddata.header['DICHFRAC'] = (
                dichroic_fraction, 'Dichroic Fraction')
            # Wavelength fit statistics
            self.action.args.ccddata.header['AVWVSIG'] = (
                geom['avwvsig'], 'Avg. bar wave sigma (Ang)')
            self.action.args.ccddata.header['SDWVSIG'] = (
                geom['sdwvsig'], 'Stdev. var wave sigma (Ang)')
            # Pixel scales
            self.action.args.ccddata.header['PXSCL'] = (
                geom['pxscl'], 'Pixel scale along slice (deg)')
            self.action.args.ccddata.header['SLSCL'] = (
                geom['slscl'], 'Pixel scale perp. to slices (deg)')
            # Geometry origins
            self.action.args.ccddata.header['CBARSNO'] = (
                geom['cbarsno'], 'Continuum bars image number')
            self.action.args.ccddata.header['CBARSFL'] = (
                geom['cbarsfl'], 'Continuum bars image filename')
            self.action.args.ccddata.header['ARCNO'] = (
                geom['arcno'], 'Arc image number')
            self.action.args.ccddata.header['ARCFL'] = (
                geom['arcfl'], 'Arc image filename')
            self.action.args.ccddata.header['GEOMFL'] = (
                geom_file.split('/')[-1], 'Geometry file')
            # WCS
            self.action.args.ccddata.header['IFUPA'] = (
                skypa, 'IFU position angle (degrees)')
            self.action.args.ccddata.header['IFUROFF'] = (
                self.config.instrument.ROTOFF, 'IFU-SKYPA offset (degrees)')
            self.action.args.ccddata.header['WCSDIM'] = (
                3, 'number of dimensions in WCS')
            self.action.args.ccddata.header['WCSNAME'] = 'KCWI'
            self.action.args.ccddata.header['EQUINOX'] = 2000.
            self.action.args.ccddata.header['RADESYS'] = 'FK5'
            self.action.args.ccddata.header['CTYPE1'] = 'RA---TAN'
            self.action.args.ccddata.header['CTYPE2'] = 'DEC--TAN'
            self.action.args.ccddata.header['CTYPE3'] = ('AWAV',
                                                         'Air Wavelengths')
            self.action.args.ccddata.header['CUNIT1'] = ('deg', 'RA units')
            self.action.args.ccddata.header['CUNIT2'] = ('deg', 'DEC units')
            self.action.args.ccddata.header['CUNIT3'] = ('Angstrom',
                                                         'Wavelength units')
            self.action.args.ccddata.header['CNAME1'] = ('KCWI RA', 'RA name')
            self.action.args.ccddata.header['CNAME2'] = ('KCWI DEC', 'DEC name')
            self.action.args.ccddata.header['CNAME3'] = ('KCWI Wavelength',
                                                         'Wavelength name')
            self.action.args.ccddata.header['CRVAL1'] = (ra, 'RA zeropoint')
            self.action.args.ccddata.header['CRVAL2'] = (dec, 'DEC zeropoint')
            self.action.args.ccddata.header['CRVAL3'] = (geom['wave0out'],
                                                         'Wavelength zeropoint')
            self.action.args.ccddata.header['CRPIX1'] = (crpix1,
                                                         'RA reference pixel')
            self.action.args.ccddata.header['CRPIX2'] = (crpix2,
                                                         'DEC reference pixel')
            self.action.args.ccddata.header['CRPIX3'] = (
                crpix3, 'Wavelength reference pixel')
            self.action.args.ccddata.header['CD1_1'] = (
                cd11, 'RA degrees per column pixel')
            self.action.args.ccddata.header['CD2_1'] = (
                cd21, 'DEC degrees per column pixel')
            self.action.args.ccddata.header['CD1_2'] = (
                cd12, 'RA degrees per row pixel')
            self.action.args.ccddata.header['CD2_2'] = (
                cd22, 'DEC degrees per row pixel')
            self.action.args.ccddata.header['CD3_3'] = (
                geom['dwout'], 'Wavelength Angstroms per pixel')
            self.action.args.ccddata.header['LONPOLE'] = (
                180.0, 'Native longitude of Celestial pole')
            self.action.args.ccddata.header['LATPOLE'] = (
                0.0, 'Native latitude of Celestial pole')
            # write out cube
            self.action.args.ccddata.header['HISTORY'] = log_string
            self.action.args.ccddata.data = out_cube
            self.action.args.ccddata.uncertainty.array = out_uube
            self.action.args.ccddata.mask = out_mube
            self.action.args.ccddata.flags = out_fube
            if data_nsk is not None:
                self.action.args.ccddata.noskysub = out_kube

            # write out int image
            kcwi_fits_writer(self.action.args.ccddata,
                             table=self.action.args.table,
                             output_file=self.action.args.name,
                             output_dir=self.config.instrument.output_directory,
                             suffix="icube")
            self.context.proctab.update_proctab(frame=self.action.args.ccddata,
                                                suffix="icube",
                                                filename=self.action.args.name)
            self.context.proctab.write_proctab(
                tfil=self.config.instrument.procfile)

            # check for obj, sky outputs
            if obj is not None:
                out_obj = CCDData(out_oube,
                                  meta=self.action.args.ccddata.header,
                                  unit=self.action.args.ccddata.unit)
                kcwi_fits_writer(
                    out_obj, output_file=self.action.args.name,
                    output_dir=self.config.instrument.output_directory,
                    suffix="ocube")
            if sky is not None:
                out_sky = CCDData(out_sube,
                                  meta=self.action.args.ccddata.header,
                                  unit=self.action.args.ccddata.unit)
                kcwi_fits_writer(
                    out_sky, output_file=self.action.args.name,
                    output_dir=self.config.instrument.output_directory,
                    suffix="scube")
            # check for dew outputs
            if dew is not None:
                out_dew = CCDData(out_dube,
                                  meta=self.action.args.ccddata.header,
                                  unit=dew.unit)
                kcwi_fits_writer(
                    out_dew, output_file=self.action.args.name,
                    output_dir=self.config.instrument.output_directory,
                    suffix="dcube")
        else:
            self.logger.error("Geometry file not found: %s" % geom_file)

        self.logger.info(log_string)

        return self.action.args
    # END: class MakeCube()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer, \
    kcwi_fits_reader, get_master_name, strip_fname
from kcwidrp.core.kcwi_correct_extin import kcwi_correct_extin

import os
import numpy as np
from scipy.interpolate import interp1d

from astropy.io import fits as pf
from astropy import units as u
from astropy.nddata import CCDData


class FluxCalibrate(BasePrimitive):
    """
    Perform flux calibration.

    Uses inverse sensitivity curve derived from MakeInvsens to flux calibrate
    input observation.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        """
        Checks if we can calibrate flux based on the processing table
        """
        self.logger.info("Checking precondition for FluxCalibrate")
        target_type = 'INVSENS'
        tab = self.context.proctab.search_proctab(
            frame=self.action.args.ccddata, target_type=target_type,
            nearest=True)
        self.logger.info("pre condition got %d invsens files, expected >= 1"
                         % len(tab))
        if len(tab) <= 0:
            self.action.args.invsname = None
            return False
        else:
            self.action.args.invsname = get_master_name(tab, target_type)
            return True

    def _perform(self):
        # Header keyword to update
        key = 'STDCOR'
        keycom = 'std corrected?'
        target_type = 'INVSENS'
        obj = None
        sky = None

        self.logger.info("Calibrating object flux")
        tab = self.context.proctab.search_proctab(
            frame=self.action.args.ccddata, target_type=target_type,
            nearest=True)
        self.logger.info("%d invsens files found" % len(tab))

        if self.action.args.invsname is not None:

            # read in master calibration (inverse sensitivity)
            invsname = self.action.args.invsname
            self.logger.info("Reading invsens: %s" % invsname)
            hdul = pf.open(os.path.join(self.config.instrument.cwd,
                                        'redux', invsname))
            mcal = hdul[0].data[1, :]
            mchdr = hdul[0].header
            hdul.close()
            # get dimensions
            mcsz = mcal.shape
            # get master std waves
            mcw0 = mchdr['CRVAL1']
            mcdw = mchdr['CDELT1']
            mcwav = mcw0 + np.arange(mcsz[0]) * mcdw
            # get master std image number
            msimgno = mchdr['FRAMENO']
            # get input object image dimensions
            sz = self.action.args.ccddata.data.shape
            # get object waves
            w0 = self.action.args.ccddata.header['CRVAL3']
            dw = self.action.args.ccddata.header['CD3_3']
            wav = w0 + np.arange(sz[0]) * dw
            # get exposure time
            expt = self.action.args.ccddata.header['XPOSURE']
            if expt <= 0:
                self.logger.warning("XPOSURE at 0.0, trying TTIME")
                expt = self.action.args.ccddata.header['TTIME']
                if expt <= 0:
                    self.logger.warning("No valid exposure time found, "
                                        "using 1s")
                    expt = 1.0
            # resample onto object waves, if needed
            if w0 != mcw0 or dw != mcdw or wav[-1] != mcwav[-1] or \
                    sz[0] != mcsz[0]:
                self.logger.warning("wavelength scales not identical, "
                                    "resampling standard")
                print(w0, mcw0, dw, mcdw, wav[-1], mcwav[-1], sz[0], mcsz[0])
                mcint = interp1d(mcwav, mcal, kind='cubic',
                                 fill_value='extrapolate')
                mscal = mcint(wav) * 1.e16 / expt
            else:
                mscal = mcal * 1.e16 / expt

            # extinction correct calibration
            kcwi_correct_extin(mscal, self.action.args.ccddata.header,
                               logger=self.logger)
            # do calibration
            for isl in range(sz[2]):
                for ix in range(sz[1]):
                    self.action.args.ccddata.data[:, ix, isl] *= mscal
                    if self.action.args.ccddata.noskysub is not None:
                        self.action.args.ccddata.noskysub[:, ix, isl] *= mscal
                    self.action.args.ccddata.uncertainty.array[:, ix, isl] *= \
                        mscal

            # check for obj, sky cubes
            if self.action.args.nasmask and self.action.args.numopen > 1:
                ofn = self.action.args.name
                # obj cube
                objfn = strip_fname(ofn) + '_ocubed.fits'
                full_path = os.path.join(
                    self.config.instrument.cwd,
                    self.config.instrument.output_directory, objfn)
                if os.path.exists(full_path):
                    obj = kcwi_fits_reader(full_path)[0]
                    # do calibration
                    for isl in range(sz[2]):
                        for ix in range(sz[1]):
                            obj.data[:, ix, isl] *= mscal
                # sky cube
                skyfn = strip_fname(ofn) + '_scubed.fits'
                full_path = os.path.join(
                    self.config.instrument.cwd,
                    self.config.instrument.output_directory, skyfn)
                if os.path.exists(full_path):
                    sky = kcwi_fits_reader(full_path)[0]
                    # do calibration
                    for isl in range(sz[2]):
                        for ix in range(sz[1]):
                            sky.data[:, ix, isl] *= mscal

            # units
            flam16_u = 1.e16 * u.erg / (u.angstrom * u.cm ** 2 * u.s)
            self.action.args.ccddata.unit = flam16_u
            self.action.args.ccddata.uncertainty.unit = flam16_u
            if obj is not None:
                obj.unit = flam16_u
            if sky is not None:
                sky.unit = flam16_u
            # update header keywords
            self.action.args.ccddata.header[key] = (True, keycom)
            self.action.args.ccddata.header['MSFILE'] = (invsname,
                                                         "Master std filename")
            self.action.args.ccddata.header['MSIMNO'] = (
                msimgno, 'master std image number')
        else:

            self.action.args.ccddata.header[key] = (False, keycom)

        log_string = FluxCalibrate.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string

        # write out icubes image
        kcwi_fits_writer(self.action.args.ccddata,
                         table=self.action.args.table,
                         output_file=self.action.args.name,
                         output_dir=self.config.instrument.output_directory,
                         suffix="icubes")
        self.context.proctab.update_proctab(frame=self.action.args.ccddata,
                                            suffix="icubes",
                                            filename=self.action.args.name)
        self.context.proctab.write_proctab(tfil=self.config.instrument.procfile)

        # check for sky, obj cube
        if obj is not None:
            out_obj = CCDData(obj, meta=self.action.args.ccddata.header,
                              unit=self.action.args.ccddata.unit)
            kcwi_fits_writer(
                out_obj, output_file=self.action.args.name,
                output_dir=self.config.instrument.output_directory,
                suffix="ocubes")

        if sky is not None:
            out_sky = CCDData(sky, meta=self.action.args.ccddata.header,
                              unit=self.action.args.ccddata.unit)
            kcwi_fits_writer(
                out_sky, output_file=self.action.args.name,
                output_dir=self.config.instrument.output_directory,
                suffix="scubes")

        self.logger.info(log_string)

        return self.action.args
    # END: class FluxCalibrate()



from keckdrpframework.primitives.base_img import BaseImg
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_reader, \
    kcwi_fits_writer, strip_fname  # , get_master_name

import os
import ccdproc
import numpy as np
from astropy.stats import mad_std


class MakeMasterArc(BaseImg):
    """
    Stack arc frames into master arc

    Generate a master arc frame based on the instrument config parameter
    arc_min_nframes, which defaults to 1 for the BLUE channel and 3 for the RED
    channel.  It is assumed that each frame is well-exposed and the combine
    method 'median' will be used to mitigate cosmic rays (especially for the RED
    channel).  A high sigma clipping of 2.0 is used to help with the CRs.

    Uses the ccdproc.combine routine to peform the stacking.

    Writes out a \*_marc.fits file and records a master arc frame in the proc
    table, no matter how many frames are combined.

    """

    def __init__(self, action, context):
        BaseImg.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        """
        Checks if we can build a stacked frame based on the processing table
        """
        # get list of arc frames
        self.logger.info("Checking precondition for stack_arcs")
        self.combine_list = self.context.proctab.search_proctab(
            frame=self.action.args.ccddata, target_type='ARCLAMP',
            target_group=self.action.args.groupid)
        self.logger.info(f"pre condition got {len(self.combine_list)},"
                         f" expecting {self.action.args.min_files}")
        # do we meet the criterion?
        if len(self.combine_list) >= self.action.args.min_files:
            return True
        else:
            return False

    def _perform(self):
        """
        Returns an Argument() with the parameters that depends on this operation
        """
        args = self.action.args
        method = 'median'
        sig_up = 2.0
        suffix = args.new_type.lower()
        log_string = MakeMasterArc.__module__

        combine_list = list(self.combine_list['filename'])
        # get master arc output name
        maname = strip_fname(combine_list[0]) + '_' + suffix + '.fits'
        if self.action.args.min_files > 1:
            stack = []
            stackf = []
            for arc in combine_list:
                # get arc intensity (int) image file name in redux directory
                stackf.append(arc.split('.fits')[0] + '_int.fits')
                arcfn = os.path.join(args.in_directory, stackf[-1])
                # using [0] gets just the image data
                stack.append(kcwi_fits_reader(arcfn)[0])

            stacked = ccdproc.combine(stack, method=method, sigma_clip=True,
                                      sigma_clip_low_thresh=None,
                                      sigma_clip_high_thresh=sig_up,
                                      sigma_clip_func=np.ma.median,
                                      sigma_clip_dev_func=mad_std)
            stacked.unit = stack[0].unit
            stacked.header['IMTYPE'] = args.new_type
            stacked.header['NSTACK'] = (len(combine_list),
                                        'number of images stacked')
            stacked.header['STCKMETH'] = (method, 'method used for stacking')
            stacked.header['STCKSIGU'] = (sig_up,
                                          'Upper sigma rejection for stacking')
            for ii, fname in enumerate(stackf):
                stacked.header['STACKF%d' % (ii + 1)] = (fname,
                                                         "stack input file")
            stacked.header['HISTORY'] = log_string
            self.action.args.ccddata = stacked

            kcwi_fits_writer(stacked, output_file=maname,
                             output_dir=self.config.instrument.output_directory)
            self.context.proctab.update_proctab(frame=stacked, suffix=suffix,
                                                newtype=args.new_type,
                                                filename=stacked.header[
                                                    'OFNAME'])
            self.action.args.name = stacked.header['OFNAME']
        else:
            self.action.args.ccddata.header['IMTYPE'] = args.new_type
            self.action.args.ccddata.header['HISTORY'] = log_string
            kcwi_fits_writer(self.action.args.ccddata, output_file=maname,
                             output_dir=self.config.instrument.output_directory)
            self.context.proctab.update_proctab(frame=self.action.args.ccddata,
                                                suffix=suffix,
                                                newtype=args.new_type,
                                                filename=self.action.args.name)
        self.logger.info(log_string)

        self.context.proctab.write_proctab(tfil=self.config.instrument.procfile)
        return self.action.args
    # END: class MakeMasterArc()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""This module corresponds to the misc directory in idlutils.
"""
import numpy as np
from . import PydlutilsException


def decode_mixed(x):
    """Convert bytes in Numpy arrays into strings.  Leave other stuff alone.

    Parameters
    ----------
    x : object
        Input object.

    Returns
    -------
    object
        If `x` has a ``decode()`` method, ``x.decode()`` will be returned.
        Otherwise `x` will be returned unchanged.
    """
    try:
        return x.decode()
    except:
        return x


def djs_laxisgen(dims, iaxis=0):
    """Returns an integer array where each element of the array is set
    equal to its index number along the specified axis.

    Parameters
    ----------
    dims : :class:`list`
        Dimensions of the array to return.
    iaxis : :class:`int`, optional
        Index along this dimension.

    Returns
    -------
    :class:`numpy.ndarray`
        An array of indexes with ``dtype=int32``.

    Raises
    ------
    :exc:`ValueError`
        If `iaxis` is greater than or equal to the number of dimensions.

    Notes
    -----
    For two or more dimensions, there is no difference between this routine
    and :func:`~pydl.pydlutils.misc.djs_laxisnum`.

    Examples
    --------
    >>> from pydl.pydlutils.misc import djs_laxisgen
    >>> print(djs_laxisgen([4,4]))
    [[0 0 0 0]
     [1 1 1 1]
     [2 2 2 2]
     [3 3 3 3]]
    """
    ndimen = len(dims)
    if ndimen == 1:
        return np.arange(dims[0], dtype='i4')
    return djs_laxisnum(dims, iaxis)


def djs_laxisnum(dims, iaxis=0):
    """Returns an integer array where each element of the array is set equal
    to its index number in the specified axis.

    Parameters
    ----------
    dims : :class:`list`
        Dimensions of the array to return.
    iaxis : :class:`int`, optional
        Index along this dimension.

    Returns
    -------
    :class:`numpy.ndarray`
        An array of indexes with ``dtype=int32``.

    Raises
    ------
    :exc:`ValueError`
        If `iaxis` is greater than or equal to the number of dimensions, or
        if number of dimensions is greater than three.

    Notes
    -----
    For two or more dimensions, there is no difference between this routine
    and :func:`~pydl.pydlutils.misc.djs_laxisgen`.

    Examples
    --------
    >>> from pydl.pydlutils.misc import djs_laxisnum
    >>> print(djs_laxisnum([4,4]))
    [[0 0 0 0]
     [1 1 1 1]
     [2 2 2 2]
     [3 3 3 3]]
    """
    ndimen = len(dims)
    result = np.zeros(dims, dtype='i4')
    if ndimen == 1:
        pass
    elif ndimen == 2:
        if iaxis == 0:
            for k in range(dims[0]):
                result[k, :] = k
        elif iaxis == 1:
            for k in range(dims[1]):
                result[:, k] = k
        else:
            raise ValueError("Bad value for iaxis: {0:d}".format(iaxis))
    elif ndimen == 3:
        if iaxis == 0:
            for k in range(dims[0]):
                result[k, :, :] = k
        elif iaxis == 1:
            for k in range(dims[1]):
                result[:, k, :] = k
        elif iaxis == 2:
            for k in range(dims[2]):
                result[:, :, k] = k
        else:
            raise ValueError("Bad value for iaxis: {0:d}".format(iaxis))
    else:
        raise ValueError("{0:d} dimensions not supported.".format(ndimen))
    return result


def struct_print(array, filename=None, formatcodes=None, alias=None,
                 fdigit=5, ddigit=7, html=False, no_head=False,
                 silent=False):
    """Print a NumPy record array (analogous to an IDL structure) in a
    nice way.

    Parameters
    ----------
    array : :class:`numpy.ndarray`
        A record array to print.
    filename : :class:`str` or file-like, optional
        If supplied, write to this file.
    formatcodes : :class:`dict`, optional
        If supplied, use explicit format for certain columns.
    alias : :class:`dict`, optional
        If supplied, use this mapping of record array column names to printed
        column names.
    fdigit : :class:`int`, optional
        Width of 32-bit floating point columns, default 5.
    ddigit : :class:`int`, optional
        Width of 64-bit floating point columns, default 7.
    html : :class:`bool`, optional
        If ``True``, print an html table.
    no_head : :class:`bool`, optional
        If ``True``, *don't* print a header line.
    silent : :class:`bool`, optional
        If ``True``, do not print the table, just return it.

    Returns
    -------
    :func:`tuple`
        A tuple containing a list of the lines in the table.  If `html` is
        ``True``, also returns a list of lines of CSS for formatting the
        table.

    Examples
    --------
    import numpy as np
    from pydl.pydlutils.misc import struct_print
    struct_print(np.array([(1,2.34,'five'),(2,3.456,'seven'),(3,4.5678,'nine')],
        dtype=[('a','i4'),('bb','f4'),('ccc','S5')]),silent=True)
    (['a bb          ccc  ', '- ----------- -----', '1        2.34 five ',
         '2       3.456 seven', '3      4.5678 nine '], [])
    """
    if html:
        headstart = '<tr><th>'
        headsep = '</th><th>'
        headend = '</th></tr>'
        colstart = '<tr><td>'
        colsep = '</td><td>'
        colend = '</td></tr>'
        css = ['<style type="text/css">',
               'table {',
               '    border-collapse: collapse;',
               '}',
               'th {',
               '    padding: 2px;',
               '    text-align: right;',
               '    border: 1px solid black;',
               '    font-weight: bold;',
               '}',
               'td {',
               '    padding: 2px;',
               '    text-align: right;',
               '    border: 1px solid black;',
               '}',
               '</style>']
    else:
        headstart = ''
        headsep = ' '
        headend = ''
        colstart = ''
        colsep = ' '
        colend = ''
        css = list()
    #
    # Alias should be a dictionary that maps structure names to column names
    #
    if alias is None:
        #
        # Create a dummy alias dictionary
        #
        alias = dict(list(zip(array.dtype.names, array.dtype.names)))
    else:
        #
        # Fill in any missing values of the alias dictionary
        #
        for tag in array.dtype.names:
            if tag not in alias:
                alias[tag] = tag
    #
    # Formatcodes allows an override for certain columns.
    #
    if formatcodes is None:
        formatcodes = dict()
    #
    # This dictionary will hold the number of characters in each column
    #
    nchar = dict()
    #
    # Construct format codes for each column
    #
    for k, tag in enumerate(array.dtype.names):
        if tag in formatcodes:
            thiscode = formatcodes[tag]
            thisn = len(thiscode.format(array[tag][0]))
        else:
            d = array.dtype.fields[tag][0]
            if d.kind == 'i' or d.kind == 'u':
                thisn = max(max(len(str(array[tag].min())),
                            len(str(array[tag].max()))), len(tag))
                thiscode = "{{{0:d}:{1:d}d}}".format(k, thisn)
            elif d.kind == 'f':
                if d.itemsize == 8:
                    prec = ddigit
                else:
                    prec = fdigit
                thisn = prec + 6
                if array[tag].min() < 0:
                    thisn += 1
                thiscode = "{{{0:d}:{1:d}.{2:d}g}}".format(k, thisn, prec)
            elif d.kind == 'S' or d.kind == 'U':
                thisn = max(d.itemsize, len(tag))
                thiscode = "{{{0:d}:{1:d}s}}".format(k, thisn)
            else:
                raise PydlutilsException(
                        "Unsupported kind: {0}".format(d.kind))
            formatcodes[tag] = thiscode
        nchar[tag] = thisn
    #
    # Start building an array of lines
    #
    lines = list()
    #
    # Construct header lines
    #
    if html:
        lines.append('<table>')
        hdr1 = (headstart + headsep.join([alias[tag]
                for tag in array.dtype.names]) + headend)
        lines.append(hdr1)
    else:
        if not no_head:
            hdr1 = (headstart + headsep.join([("{{0:{0:d}s}}".format(
                    nchar[tag])).format(alias[tag])
                    for tag in array.dtype.names]) + headend)
            hdr2 = (headstart + headsep.join(['-' * nchar[tag]
                    for tag in array.dtype.names]) + headend)
            lines.append(hdr1)
            lines.append(hdr2)
    #
    # Create a format string for the data from the individual format codes
    #
    rowformat = (colstart + colsep.join([formatcodes[tag]
                 for tag in array.dtype.names]) + colend)
    for k in range(array.size):
        lines.append(rowformat.format(
                    *([decode_mixed(i) for i in array[k].tolist()])))
    if html:
        lines.append('</table>')
    f = None   # This variable will store a file handle
    close_file = False
    if filename is not None:
        if hasattr(filename, 'write'):
            f = filename
        else:
            f = open(filename, 'w+b')
            close_file = True
    if f is None:
        if not silent:  # pragma: no cover
            print("\n".join(lines)+"\n")
    else:
        f.write(("\n".join(lines)+"\n").encode('utf-8'))
        if close_file:
            f.close()
    return lines, css



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.core.bokeh_plotting import bokeh_plot
from kcwidrp.core.kcwi_plotting import save_plot
from kcwidrp.primitives.kcwi_file_primitives import plotlabel

from bokeh.plotting import figure
from bokeh.models import Range1d
import numpy as np
import scipy as sp
from scipy.interpolate import interpolate
from scipy.signal.windows import boxcar
from scipy.optimize import curve_fit
from scipy.stats import sigmaclip
import time
import os
import traceback


def gaus(x, a, mu, sigma):
    """
    Gaussian fitting function.

    Args:
        x (float): x value
        a (float): amplitude
        mu (float): x position of center of Gaussian
        sigma (float): Gaussian width

    Returns:
        Gaussian value at x position

    """
    return a * np.exp(-(x - mu) ** 2 / (2. * sigma ** 2))


def get_line_window(y, c, thresh=0., logger=None, strict=False, maxwin=100,
                    frac_max=0.5):
    """
    Find a window that includes the specified amount of the line.

    Get a useful window on a given line to allow finding an accurate peak for
    each line.  Start at input center and iterate above and below until the
    window encompasses the specified fraction of the line maximum flux.
    Rejects lines that are problematic.

    Args:
        y (list of floats): flux values
        c (float): the nominal center of the line in pixels
        thresh (float): threshhold for rejection (default: 0.)
        logger (log object): logging instance for output
        strict (bool): should we apply strict criterion? (default: False)
        maxwin (int): largest possible window in pixels (default: 100)
        frac_max (float): fraction of maximum flux to encompass (default: 0.5)

    :returns:
        - x0 (int): lower limit of window in pixels
        - x1 (int): upper limit of window in pixels
        - count (int): number of pixels in window

    """
    verbose = logger is not None
    nx = len(y)
    # check edges
    if c < 2 or c > nx - 2:
        if verbose:
            logger.info("input center too close to edge")
        return None, None, 0
    # get initial values
    x0 = c - 2
    x1 = c + 2
    mx = np.nanmax(y[x0:x1+1])
    count = 5
    # check low side
    if x0 - 1 < 0:
        if verbose:
            logger.info("max check: low edge hit")
        return None, None, 0
    while y[x0-1] > mx:
        x0 -= 1
        count += 1
        if x0 - 1 < 0:
            if verbose:
                logger.info("Max check: low edge hit")
            return None, None, 0

    # check high side
    if x1 + 1 >= nx:
        if verbose:
            logger.info("max check: high edge hit")
        return None, None, 0
    while y[x1+1] > mx:
        x1 += 1
        count += 1
        if x1 + 1 >= nx:
            if verbose:
                logger.info("Max check: high edge hit")
            return None, None, 0
    # how big is our window?
    if (x1 - x0) > maxwin:
        if verbose:
            logger.info("Window expanded beyond limit")
        return None, None, 0
    # adjust starting window to center on max
    cmx = x0 + y[x0:x1+1].argmax()
    x0 = cmx - 2
    x1 = cmx + 2
    mx = np.nanmax(y[x0:x1 + 1])
    # make sure max is high enough
    if mx < thresh:
        return None, None, 0
    #
    # expand until we get to max_frac
    fmx = mx * frac_max
    #
    # Low index side
    prev = mx
    while y[x0] > fmx:
        if y[x0] > mx or x0 <= 0 or y[x0] > prev:
            if verbose:
                if y[x0] > mx:
                    logger.info("frac_max check: low index err - missed max")
                if x0 <= 0:
                    logger.info("frac_max check: low index err - at edge")
                if y[x0] > prev:
                    logger.info("frac_max check: low index err - wiggly")
            return None, None, 0
        prev = y[x0]
        x0 -= 1
        count += 1
    # High index side
    prev = mx
    while y[x1] > fmx:
        if y[x1] > mx or x1 >= nx or y[x1] > prev:
            if verbose:
                if y[x1] > mx:
                    logger.info("frac_max check: high index err - missed max")
                if x1 >= nx:
                    logger.info("frac_max check: high index err - at edge")
                if y[x1] > prev:
                    logger.info("frac_max check: high index err - wiggly")
            return None, None, 0
        prev = y[x1]
        if x1 < (nx-1):
            x1 += 1
            count += 1
        else:
            if verbose:
                logger.info("Edge encountered")
            return None, None, 0
    if strict:
        # where did we end up?
        if c < x0 or x1 < c:
            if verbose:
                logger.info("initial position outside final window")
            return None, None, 0

    return x0, x1, count
    # END: get_line_window()


def findpeaks(x, y, wid, sth, ath, pkg=None, verbose=False):
    """
    Find peaks in spectrum

    Uses a gradient to determine where peaks are and then performs some
    cleaning of blended lines using sigma rejection.

    Args:
        x (list of floats): x values of vector
        y (list of floats): y values of vector
        wid (int): boxcar smoothing width in pixels
        sth (float): slope threshhold
        ath (float): amplitude threshhold
        pkg (float): limit on peak wandering in pixels
        verbose (bool): control output messages

    :returns:
        - list of peaks
        - average sigma of cleaned peaks
        - list of y values at peaks

    """
    # derivative
    grad = np.gradient(y)
    # smooth derivative
    win = boxcar(wid)
    d = sp.signal.convolve(grad, win, mode='same') / sum(win)
    # size
    nx = len(x)
    # set up windowing
    if not pkg:
        pkg = wid
    hgrp = int(pkg/2)
    hgt = []
    pks = []
    sgs = []
    # loop over spectrum
    # limits to avoid edges given pkg
    for i in np.arange(pkg, (nx - pkg)):
        # find zero crossings
        if np.sign(d[i]) > np.sign(d[i+1]):
            # pass slope threshhold?
            if (d[i] - d[i+1]) > sth * y[i]:
                # pass amplitude threshhold?
                if y[i] > ath or y[i+1] > ath:
                    # get subvectors around peak in window
                    xx = x[(i-hgrp):(i+hgrp+1)]
                    yy = y[(i-hgrp):(i+hgrp+1)]
                    if len(yy) > 3:
                        try:
                            # gaussian fit
                            res, _ = curve_fit(gaus, xx, yy,
                                               p0=[y[i], x[i], 1.])
                            # check offset of fit from initial peak
                            r = abs(x - res[1])
                            t = r.argmin()
                            if abs(i - t) > pkg:
                                if verbose:
                                    print(i, t, x[i], res[1], x[t])
                            else:
                                hgt.append(res[0])
                                pks.append(res[1])
                                sgs.append(abs(res[2]))
                        except RuntimeError:
                            continue
    # clean by sigmas
    cvals = []
    cpks = []
    sgmn = None
    if len(pks) > 0:
        cln_sgs, low, upp = sigmaclip(sgs, low=3., high=3.)
        for i in range(len(pks)):
            # clean only blends (overly wide lines)
            if sgs[i] < upp:
                cpks.append(pks[i])
                cvals.append(hgt[i])
        sgmn = cln_sgs.mean()
        # sgmd = float(np.nanmedian(cln_sgs))
    else:
        print("No peaks found!")
    return cpks, sgmn, cvals
    # END: findpeaks()


class GetAtlasLines(BasePrimitive):
    """
    Get relevant atlas line positions and wavelengths

    Generates a list of well observed atlas lines for generating the final
    wavelength solution.

    Uses the following configuration parameters:

    * FRACMAX: fraction of max to use for window for fitting (defaults to 0.5)
    * LINELIST: an input line list to use instead of determining one on the fly

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger
        self.action.args.atminrow = None
        self.action.args.atmaxrow = None
        self.action.args.atminwave = None
        self.action.args.atmaxwave = None
        self.action.args.at_wave = None
        self.action.args.at_flux = None

    def _pre_condition(self):
        self.logger.info("Checking for master arc")
        if 'MARC' in self.action.args.ccddata.header['IMTYPE']:
            return True
        else:
            return False

    def _perform(self):
        """Get atlas line positions for wavelength fitting"""
        self.logger.info("Finding isolated atlas lines")
        verbose = (self.config.instrument.verbose > 1)
        frac_max = self.config.instrument.FRACMAX
        self.logger.info("Finding line windows using fraction of line max of "
                         "%.2f" % frac_max)

        # get atlas wavelength range
        # get pixel values (no longer centered in the middle)
        specsz = len(self.context.arcs[self.config.instrument.REFBAR])
        xvals = np.arange(0, specsz)
        # min, max rows, trimming the ends
        minrow = 50
        maxrow = specsz - 50
        # wavelength range
        mnwvs = []
        mxwvs = []
        refbar_disp = 1.
        # Get wavelengths for each bar
        for b in range(self.config.instrument.NBARS):
            waves = np.polyval(self.action.args.twkcoeff[b], xvals)
            mnwvs.append(np.min(waves))
            mxwvs.append(np.max(waves))
            if b == self.config.instrument.REFBAR:
                refbar_disp = self.action.args.twkcoeff[b][-2]
        self.logger.info("Ref bar (%d) dispersion = %.3f Ang/px" %
                         (self.config.instrument.REFBAR, refbar_disp))
        # Get extrema (trim ends a bit)
        minwav = min(mnwvs) + 10.
        maxwav = max(mxwvs) - 10.
        wave_range = maxwav - minwav
        # Do we have a dichroic?
        if self.action.args.dich:
            if self.action.args.camera == 0:  # Blue
                maxwav = min([maxwav, 5620.])
            elif self.action.args.camera == 1:  # Red
                minwav = max([minwav, 5580.])
            else:
                self.logger.error("Camera keyword not defined!")
        dichroic_fraction = (maxwav - minwav) / wave_range
        # Get corresponding atlas range
        minrw = [i for i, v in enumerate(self.action.args.refwave)
                 if v >= minwav][0]
        maxrw = [i for i, v in enumerate(self.action.args.refwave)
                 if v <= maxwav][-1]
        self.logger.info("Min, Max wave (A): %.2f, %.2f" % (minwav, maxwav))
        if self.action.args.dich:
            self.logger.info("Dichroic fraction: %.3f" % dichroic_fraction)
        # store atlas ranges
        self.action.args.atminrow = minrw
        self.action.args.atmaxrow = maxrw
        self.action.args.atminwave = minwav
        self.action.args.atmaxwave = maxwav
        self.action.args.dichroic_fraction = dichroic_fraction
        # output filename stub
        atfnam = "arc_%05d_%s_%s_%s_atlines" % \
                 (self.action.args.ccddata.header['FRAMENO'],
                  self.action.args.illum, self.action.args.grating,
                  self.action.args.ifuname)
        # check if line list was given on command line
        if self.config.instrument.LINELIST:
            with open(self.config.instrument.LINELIST) as llfn:
                atlines = llfn.readlines()
            refws = []
            refas = []
            for line in atlines:
                if '#' in line:
                    continue
                refws.append(float(line.split()[0]))
                refas.append(float(line.split()[1]))
            self.logger.info("Read %d lines from %s" %
                             (len(refws), self.config.instrument.LINELIST))
        else:
            # get atlas sub spectrum
            atspec = self.action.args.reflux[minrw:maxrw]
            atwave = self.action.args.refwave[minrw:maxrw]
            # get ref bar arc spectrum, pixel values, and prelim wavelengths
            subxvals = xvals[minrow:maxrow]
            subyvals = self.context.arcs[self.config.instrument.REFBAR][
                       minrow:maxrow].copy()
            subwvals = np.polyval(
                self.action.args.twkcoeff[self.config.instrument.REFBAR],
                subxvals)
            # smooth subyvals
            win = boxcar(3)
            subyvals = sp.signal.convolve(subyvals, win, mode='same') / sum(win)
            # find good peaks in arc spectrum
            smooth_width = 4  # in pixels
            # peak width
            peak_width = int(self.action.args.atsig/abs(refbar_disp))
            if peak_width < 4:
                peak_width = 4
            # slope threshold
            slope_thresh = 0.7 * smooth_width / 2. / 100.
            # slope_thresh = 0.7 * smooth_width / 1000.   # more severe for arc
            # slope_thresh = 0.016 / peak_width
            # get amplitude threshold
            ampl_thresh = 0.
            self.logger.info("Using a peak_width of %d px, a slope_thresh of "
                             "%.5f a smooth_width of %d and an ampl_thresh of "
                             "%.3f" % (peak_width, slope_thresh, smooth_width,
                                       ampl_thresh))
            arc_cent, avwsg, arc_hgt = findpeaks(subwvals, subyvals,
                                                 smooth_width, slope_thresh,
                                                 ampl_thresh,
                                                 peak_width)
            avwfwhm = avwsg * 2.354
            self.logger.info("Found %d lines with <sig> = %.3f (A),"
                             " <FWHM> = %.3f (A)" % (len(arc_cent), avwsg,
                                                     avwfwhm))
            # fitting window based on grating type
            if 'H' in self.action.args.grating or \
                    'M' in self.action.args.grating:
                fwid = avwfwhm
            else:
                fwid = avwsg
            # clean near neighbors
            spec_cent = arc_cent
            spec_hgt = arc_hgt
            #
            # generate an atlas line list
            refws = []      # atlas line wavelength
            refas = []      # atlas line amplitude
            rej_win_w = []  # win rejected atlas line wavelength
            rej_win_a = []  # win rejected atlas line amplitude
            rej_fit_w = []  # fit rejected atlas line wavelength
            rej_fit_a = []  # fit rejected atlas line amplitude
            rej_par_w = []  # par rejected atlas line wavelength
            rej_par_a = []  # par rejected atlas line amplitude
            rej_dup_w = []  # dup rejected atlas line wavelength
            rej_dup_a = []  # dup rejected atlas line wavelength
            rej_fnt_w = []  # fnt rejected atlas line wavelength
            rej_fnt_a = []  # fnt rejected atlas line wavelength
            nrej = 0
            # look at each arc spectrum line
            for i, pk in enumerate(spec_cent):
                if pk <= minwav or pk >= maxwav:
                    continue
                # get atlas pixel position corresponding to arc line
                try:
                    line_x = [ii for ii, v in enumerate(atwave) if v >= pk][0]
                    # get window around atlas line to fit
                    minow, maxow, count = get_line_window(
                        atspec, line_x, logger=(self.logger if verbose
                                                else None),
                        frac_max=frac_max)
                except IndexError:
                    count = 0
                    minow = None
                    maxow = None
                    self.logger.warning("line at edge: %d, %.2f, %.f2f" %
                                        (i, pk, max(atwave)))
                # is resulting window large enough for fitting?
                if count < 5 or not minow or not maxow:
                    # keep track of fit rejected lines
                    rej_win_w.append(pk)
                    rej_win_a.append(spec_hgt[i])
                    nrej += 1
                    self.logger.info("Atlas window rejected for line %.3f" % pk)
                    continue
                # get data to fit
                yvec = atspec[minow:maxow + 1]
                xvec = atwave[minow:maxow + 1]
                # attempt Gaussian fit
                try:
                    fit, _ = curve_fit(gaus, xvec, yvec,
                                       p0=[spec_hgt[i], pk, 1.],
                                       maxfev=5000)
                except RuntimeError as e:
                    # keep track of Gaussian fit rejected lines
                    rej_fit_w.append(pk)
                    rej_fit_a.append(spec_hgt[i])
                    nrej += 1
                    self.logger.info("Atlas Gaussian fit rejected for line "
                                     "%.3f" % pk)
                    if verbose:
                        tb_str = traceback.format_exception(etype=type(e),
                                                            value=e,
                                                            tb=e.__traceback__)
                        self.logger.info("".join(tb_str))
                    continue
                # get interpolation function of atlas line
                int_line = interpolate.interp1d(xvec, yvec, kind='cubic',
                                                bounds_error=False,
                                                fill_value='extrapolate')
                # use very dense pixel sampling
                x_dense = np.linspace(min(xvec), max(xvec), num=1000)
                # resample line with dense sampling
                y_dense = int_line(x_dense)
                # get peak amplitude and wavelength
                pki = y_dense.argmax()
                pkw = x_dense[pki]
                # calculate some diagnostic parameters for the line
                # how many atlas pixels have we moved?
                xoff = abs(pkw - fit[1]) / self.action.args.refdisp
                # what is the wavelength offset in Angstroms?
                woff = abs(pkw - pk)
                # what fraction of the canonical fit width is the line?
                wrat = abs(fit[2]) / fwid  # can be neg or pos
                # current criteria for these diagnostic parameters
                if woff > 5. or xoff > 1.5 or wrat > 1.1:
                    # keep track of par rejected atlas lines
                    rej_par_w.append(pk)
                    rej_par_a.append(spec_hgt[i])
                    nrej += 1
                    self.logger.info("Atlas line parameters rejected for line "
                                     "%.3f" % pk)
                    self.logger.info("woff = %.3f, xoff = %.2f, wrat = %.3f" %
                                     (woff, xoff, wrat))
                    continue
                # check for duplicated lines
                if pkw in refws:
                    rej_dup_w.append(pk)
                    rej_dup_a.append(spec_hgt[i])
                    nrej += 1
                    self.logger.info("Atlas line duplicated for line %.3f" % pk)
                    continue
                self.logger.info("Atlas line accepted: %.3f" % pkw)
                refws.append(pkw)
                refas.append(y_dense[pki])
            # eliminate faintest lines if we have a large number
            self.logger.info("number of remaining lines: %d" % len(refas))
            if len(refas) > 400:
                # sort on flux
                sf = np.argsort(refas)
                refws = np.asarray(refws)[sf]
                refas = np.asarray(refas)[sf]
                # remove faintest two-thirds
                hlim = int(len(refas) * 0.67)
                refws = refws[hlim:]
                refas = refas[hlim:]
                rej_fnt_w = refws[:hlim]
                rej_fnt_a = refas[:hlim]
                # sort back onto wavelength
                sw = np.argsort(refws)
                refws = refws[sw].tolist()
                refas = refas[sw].tolist()

            self.logger.info("Using %d generated lines" % len(refws))

            # plot final list of Atlas lines and show rejections
            norm_fac = np.nanmax(atspec)
            if self.config.instrument.plot_level >= 1:
                p = figure(title=plotlabel(self.action.args) +
                           "ATLAS LINES Ngood = %d, Nrej = %d" % (len(refws),
                                                                  nrej),
                           x_axis_label="Wavelength (A)",
                           y_axis_label="Normalized Flux",
                           plot_width=self.config.instrument.plot_width,
                           plot_height=self.config.instrument.plot_height)
                p.line(subwvals, subyvals / np.nanmax(subyvals),
                       legend_label='RefArc', color='lightgray')
                p.line(atwave, atspec / norm_fac, legend_label='Atlas',
                       color='blue')
                # Rejected: window bad
                if len(rej_win_w) > 0:
                    p.diamond(rej_win_w, rej_win_a / norm_fac,
                              legend_label='WinRej', color='cyan', size=8)
                # Rejected: fit failure
                if len(rej_fit_w) > 0:
                    p.diamond(rej_fit_w, rej_fit_a / norm_fac,
                              legend_label='FitRej', color='red', size=8)
                # Rejected: line parameter outside range
                if len(rej_par_w) > 0:
                    p.diamond(rej_par_w, rej_par_a / norm_fac,
                              legend_label='ParRej', color='orange', size=8)
                # Rejected: duplicated line
                if len(rej_dup_w) > 0:
                    p.diamond(rej_dup_w, rej_dup_a / norm_fac,
                              legend_label='DupRej', color='brown', size=8)
                # Rejected: faint
                if len(rej_fnt_w) > 0:
                    p.diamond(rej_fnt_w, rej_fnt_a / norm_fac,
                              legend_label='FntRej', color='magenta', size=8)
                # Kept
                p.diamond(refws, refas / norm_fac, legend_label='Kept',
                          color='green', size=16)
                p.line([minwav, minwav], [-0.1, 1.1], legend_label='WavLim',
                       color='brown')
                p.line([maxwav, maxwav], [-0.1, 1.1], color='brown')
                p.x_range = Range1d(min([min(subwvals), minwav-10.]),
                                    max(subwvals))
                p.y_range = Range1d(-0.04, 1.04)
                bokeh_plot(p, self.context.bokeh_session)
                if self.config.instrument.plot_level >= 2:
                    input("Next? <cr>: ")
                else:
                    time.sleep(self.config.instrument.plot_pause)
                save_plot(p, filename=atfnam+".png")

        # output directory
        output_dir = os.path.join(self.config.instrument.cwd,
                                  self.config.instrument.output_directory)
        # write out final atlas line list
        atlines = np.array([refws, refas])
        atlines = atlines.T
        with open(os.path.join(output_dir, atfnam + '.txt'), 'w') as atlfn:
            np.savetxt(atlfn, atlines, fmt=['%12.3f', '%12.3f'])
        # store wavelengths, fluxes
        self.action.args.at_wave = refws
        self.action.args.at_flux = refas
        self.logger.info("Final atlas list has %d lines" % len(refws))

        log_string = GetAtlasLines.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        return self.action.args
    # END: class GetAtlasLines()



import numpy as np
from bokeh.io import export_png

import os
import logging

logger = logging.getLogger('KCWI')


def get_plot_lims(data, padding=0.05, clip=True):
    """Get plot limits using data range plus padding fraction"""
    dmin = np.nanmin(data)
    dmax = np.nanmax(data)
    ddel = np.abs(dmax - dmin)
    if ddel < 0.0001 and clip:
        ddel = 1.0
    return dmin - padding * ddel, dmax + padding * ddel


def oplot_slices(fig, yrange):
    """Overplot slices vertically on plot"""
    for ix in range(1, 24):
        sx = ix * 5 - 0.5
        fig.line([sx, sx], yrange, color='black', line_dash='dashdot')


def set_plot_lims(fig, xlim=None, ylim=None):
    """Set bokeh figure plot ranges"""
    if xlim:
        fig.x_range.start = xlim[0]
        fig.x_range.end = xlim[1]
    if ylim:
        fig.y_range.start = ylim[0]
        fig.y_range.end = ylim[1]


def save_plot(fig, filename=None):
    if filename is None:
        fnam = os.path.join('plots', 'kcwi_drp_plot.png')
    else:
        fnam = os.path.join('plots', filename)
    export_png(fig, filename=fnam)

    logger.info(">>> Saving to %s" % fnam)



"""
Created on May 05, 2021

@author: skwok, mbrodheim
"""

from keckdrpframework.core.framework import Framework
from keckdrpframework.config.framework_config import ConfigClass
from keckdrpframework.models.arguments import Arguments
from keckdrpframework.utils.drpf_logger import getLogger
from kcwidrp.core.bokeh_plotting import check_running_process
from kcwidrp.core.kcwi_get_std import is_file_kcwi_std

import subprocess
import time
import argparse
import sys
import traceback
import os
import pkg_resources

from kcwidrp.pipelines.keck_rti_pipeline import Keck_RTI_Pipeline
from kcwidrp.core.kcwi_proctab import Proctab
import logging.config


def _parse_arguments(in_args: list) -> argparse.Namespace:
    description = "KCWI pipeline CLI"

    # this is a simple case where we provide a frame and a configuration file
    parser = argparse.ArgumentParser(prog=f"{in_args[0]}",
                                     description=description)
    parser.add_argument('-c', '--config', dest="kcwi_config_file", type=str,
                        help="KCWI configuration file", default=None)
    parser.add_argument('-f', '--frames', nargs='*', type=str,
                        help='input image files (full path, list ok)',
                        default=None)
    parser.add_argument('-l', '--list', dest='file_list',
                        help='File containing a list of files to be processed',
                        default=None)
    parser.add_argument('-g', '--groups', dest='group_mode',
                        help='Use group mode: separate files by image type and '
                             'reduce in the correct order',
                        default=False, action='store_true')
    parser.add_argument('-t', '--taperfrac', dest='taperfrac', type=float,
                        help='Taper fraction for wavelength fitting',
                        default=None)
    parser.add_argument('-a', '--atlas_line_list', dest='atlas_line_list',
                        type=str, help="Atlas line list file", default=None)

    # in this case, we are loading an entire directory,
    # and ingesting all the files in that directory
    parser.add_argument('-i', '--infiles', dest="infiles",
                        help="Input files, or pattern to match", nargs="?")
    parser.add_argument('-d', '--directory', dest="dirname", type=str,
                        help="Input directory", nargs='?', default=None)
    # after ingesting the files,
    # do we want to continue monitoring the directory?
    parser.add_argument('-m', '--monitor', dest="monitor",
                        help='Continue monitoring the directory '
                             'after the initial ingestion',
                        action='store_true', default=False)

    # special arguments, ignore
    parser.add_argument("-I", "--ingest_data_only", dest="ingest_data_only",
                        action="store_true",
                        help="Ingest data and terminate")
    parser.add_argument("-w", "--wait_for_event", dest="wait_for_event",
                        action="store_true", help="Wait for events")
    parser.add_argument("-W", "--continue", dest="continuous",
                        action="store_true",
                        help="Continue processing, wait for ever")
    parser.add_argument("-s", "--start_queue_manager_only",
                        dest="queue_manager_only", action="store_true",
                        help="Starts queue manager only, no processing",)

    # kcwi specific parameter
    parser.add_argument("-p", "--proctab", dest='proctab', help='Proctab file',
                        default='kcwi.proc')

    out_args = parser.parse_args(in_args[1:])
    return out_args


def check_directory(directory):
    if not os.path.isdir(directory):
        os.makedirs(directory)
        print("Directory %s has been created" % directory)


def main():

    def process_subset(in_subset):
        for in_frame in in_subset.index:
            arguments = Arguments(name=in_frame)
            framework.append_event('next_file', arguments, recurrent=True)
    
    def process_list(in_list):
        for in_frame in in_list:
            arguments = Arguments(name=in_frame)
            framework.append_event('next_file', arguments, recurrent=True)

    args = _parse_arguments(sys.argv)

    # START HANDLING OF CONFIGURATION FILES ##########
    pkg = 'kcwidrp'

    # check for the logs diretory
    check_directory("logs")
    # check for the plots directory
    check_directory("plots")

    framework_config_file = "configs/framework.cfg"
    framework_config_fullpath = \
        pkg_resources.resource_filename(pkg, framework_config_file)

    framework_logcfg_file = 'configs/logger.cfg'
    framework_logcfg_fullpath = \
        pkg_resources.resource_filename(pkg, framework_logcfg_file)

    # add kcwi specific config files # make changes here to allow this file
    # to be loaded from the command line
    if args.kcwi_config_file is None:
        kcwi_config_file = 'configs/kcwi_koarti.cfg'
        kcwi_config_fullpath = pkg_resources.resource_filename(
            pkg, kcwi_config_file)
        kcwi_config = ConfigClass(kcwi_config_fullpath, default_section='KCWI')
    else:
        # kcwi_config_fullpath = os.path.abspath(args.kcwi_config_file)
        kcwi_config = ConfigClass(args.kcwi_config_file, default_section='KCWI')

    rti_config_file = "configs/rti.cfg"
    rti_config_fullpath = pkg_resources.resource_filename(pkg, rti_config_file)
    rti_config = ConfigClass(rti_config_fullpath, default_section='RTI')
    # END HANDLING OF CONFIGURATION FILES ##########

    # Add current working directory to config info
    kcwi_config.cwd = os.getcwd()

    # check for the output directory
    check_directory(kcwi_config.output_directory)

    try:
        framework = Framework(Keck_RTI_Pipeline, framework_config_fullpath)
        # add this line ONLY if you are using a local logging config file
        logging.config.fileConfig(framework_logcfg_fullpath)
        framework.config.instrument = kcwi_config
        framework.config.rti = rti_config
    except Exception as e:
        print("Failed to initialize framework, exiting ...", e)
        traceback.print_exc()
        sys.exit(1)
    framework.context.pipeline_logger = getLogger(framework_logcfg_fullpath,
                                                  name="KCWI")
    framework.logger = getLogger(framework_logcfg_fullpath,
                                 name="DRPF")

    if args.infiles is not None:
        framework.config.file_type = args.infiles

    # check for taperfrac argument
    if args.taperfrac:
        def_tf = getattr(framework.config.instrument, 'TAPERFRAC', None)
        if def_tf is not None:
            framework.context.pipeline_logger.info(
                "Setting new taperfrac = %.3f" % args.taperfrac)
            framework.config.instrument.TAPERFRAC = args.taperfrac

    # check for atlas line list argument
    if args.atlas_line_list:
        def_ll = getattr(framework.config.instrument, 'LINELIST', None)
        if def_ll is not None:
            framework.context.pipeline_logger.info(
                "Using line list %s instead of generated list" %
                args.atlas_line_list)
            framework.config.instrument.LINELIST = args.atlas_line_list

    # start the bokeh server is requested by the configuration parameters
    if framework.config.instrument.enable_bokeh is True:
        if check_running_process(process='bokeh') is False:
            subprocess.Popen('bokeh serve', shell=True)
            # --session-ids=unsigned --session-token-expiration=86400',
            # shell=True)
            time.sleep(5)
        # subprocess.Popen('open http://localhost:5006?bokeh-session-id=kcwi',
        # shell=True)

    # initialize the proctab and read it
    framework.context.proctab = Proctab()
    framework.context.proctab.read_proctab(tfil=args.proctab)

    framework.logger.info("Framework initialized")
    framework.logger.info(f"RTI url is {framework.config.rti.rti_url}")

    # add a start_bokeh event to the processing queue,
    # if requested by the configuration parameters
    if framework.config.instrument.enable_bokeh:
        framework.append_event('start_bokeh', None)

    # important: to be able to use the grouping mode, we need to reset
    # the default ingestion action to no-event,
    # otherwise the system will automatically trigger the next_file event
    # which initiates the processing
    if args.group_mode is True:
        # set the default ingestion event to None
        framework.config.default_ingestion_event = "add_only"

    # start queue manager only (useful for RPC)
    if args.queue_manager_only:
        # The queue manager runs for ever.
        framework.logger.info("Starting queue manager only, no processing")
        framework.start_queue_manager()

    # in the next two ingest_data command, if we are using standard mode,
    # the first event generated is next_file.
    # if we are in groups mode (aka smart mode), then the first event generated
    # is no_event then, manually, a next_file event is generated for each group
    # specified in the variable imtypes

    # single frame processing
    elif args.frames:
        framework.ingest_data(None, args.frames, False)

    # processing of a list of files contained in a file
    elif args.file_list:
        frames = []
        with open(args.file_list) as file_list:
            for frame in file_list:
                if "#" not in frame:
                    frames.append(frame.strip('\n'))
        framework.ingest_data(None, frames, False)

    # ingest an entire directory, trigger "next_file" (which is an option
    # specified in the config file) on each file,
    # optionally continue to monitor if -m is specified
    elif args.dirname is not None:

        framework.ingest_data(args.dirname, None, args.monitor)

    # implement the group mode
    if args.group_mode is True:
        data_set = framework.context.data_set

        # remove focus images and ccd clearing images from the dataset
        focus_frames = data_set.data_table[data_set.data_table.OBJECT ==
                                           "focus"].index
        ccdclear_frames = data_set.data_table[data_set.data_table.OBJECT ==
                                              "Clearing ccd"].index
        data_set.data_table.drop(focus_frames, inplace=True)
        data_set.data_table.drop(ccdclear_frames, inplace=True)

        # processing
        imtypes = ['BIAS', 'CONTBARS', 'ARCLAMP', 'FLATLAMP', 'DOMEFLAT', 'TWIFLAT', 'OBJECT']

        for imtype in imtypes:
            subset = data_set.data_table[
                framework.context.data_set.data_table.IMTYPE == imtype]
            if 'OBJECT' in imtype: # Ensure that standards are processed first
                object_order = []
                standard_order = []
                for frame in subset.index:
                    if is_file_kcwi_std(frame, logger=framework.context.logger):
                        standard_order.append(frame)
                    else:
                        object_order.append(frame)
                order = standard_order + object_order # Standards first
                process_list(order)
            else:
                process_subset(subset)

    framework.start(args.queue_manager_only, args.ingest_data_only,
                    args.wait_for_event, args.continuous)


if __name__ == "__main__":
    main()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.core.bokeh_plotting import bokeh_plot
from kcwidrp.core.kcwi_correct_extin import kcwi_correct_extin
from kcwidrp.core.kcwi_get_std import kcwi_get_std
from kcwidrp.core.kcwi_plotting import set_plot_lims, save_plot
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_writer, \
    kcwi_fits_reader, strip_fname

from bokeh.plotting import figure, ColumnDataSource
from scipy.signal import find_peaks
from scipy.interpolate import interp1d
from astropy.io import fits as pf
from astropy.nddata import CCDData
from astropy import units as u

import time
import os
import numpy as np


class MakeInvsens(BasePrimitive):
    """
    Generate inverse sensitivity curve from a standard star observation.

    Uses object name to determine if the observation is a standard star.  Then
    checks that the image has been processed through DAR correction.  Reads
    in the reference spectrum and compares it with the observed spectrum to
    generate the inverse sensitivity curve that can be used to flux calibrate
    science observations.

    Outputs FITS inverse sensitivity spectrum along with diagnostic plots that
    include residuals between reference spectrum and calibrated observed
    spectrum and effective area and efficiency plots as a function of
    wavelength.

    """
    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        """Checks if object is a standard star"""
        self.logger.info("Checking precondition for MakeInvsens")

        stdfile = None
        stdname = None
        if 'object' in self.action.args.imtype.lower():
            self.logger.info("Checking OBJECT keyword")
            stdfile, stdname = kcwi_get_std(
                self.action.args.ccddata.header['OBJECT'], self.logger)
            if not stdfile:
                self.logger.info("Checking TARGNAME keyword")
                stdfile, stdname = kcwi_get_std(
                    self.action.args.ccddata.header['TARGNAME'], self.logger)
        else:
            self.logger.warning("Not object type: %s" %
                                self.action.args.imtype)
        self.action.args.stdfile = stdfile
        self.action.args.stdname = stdname
        # have we been processed correctly?
        if 'DARCOR' in self.action.args.ccddata.header:
            if self.action.args.ccddata.header['DARCOR']:
                # does the standard file exist?
                if self.action.args.stdfile is not None:
                    # have we been stacked?
                    if 'NSTACK' in self.action.args.ccddata.header:
                        nstack = self.action.args.ccddata.header['NSTACK']
                    else:
                        nstack = 0
                    # does file already exist?
                    ofn = self.action.args.name
                    msname = strip_fname(ofn) + '_invsens.fits'
                    rdir = self.config.instrument.output_directory
                    invsensf = os.path.join(self.config.instrument.cwd,
                                            rdir,
                                            msname)
                    if os.path.exists(invsensf):
                        if nstack <= 0:
                            self.logger.warning("Master cal already exists: %s"
                                                % invsensf)
                            return False
                        else:
                            os.unlink(invsensf)
                            self.logger.info("Master cal will re-generated "
                                             "from stacked image")
                            return True
                    else:
                        self.logger.info("Master cal will be generated.")
                        return True
                else:
                    self.logger.warning("Not a KCWI standard star observation.")
                    return False
            else:
                self.logger.warning("DAR not corrected, cannot generate "
                                    "inverse sensitivity")
        else:
            self.logger.warning("Not processed enough to generate "
                                "inverse sensitivity")
            return False

    def _perform(self):
        self.logger.info("Making inverse sensitivity curve")

        suffix = 'invsens'
        stdname = self.action.args.stdname

        do_plots = self.config.instrument.plot_level >= 3

        # get size
        sz = self.action.args.ccddata.data.shape
        # default pixel ranges
        z = np.arange(sz[0])
        z0 = 175
        z1 = sz[0] - 175
        # get exposure time
        expt = self.action.args.ccddata.header['XPOSURE']
        if expt == 0.:
            self.logger.warning("No exposure time found, setting to 1s")
            expt = 1.
        else:
            self.logger.info("Using exposure time of %.1f" % expt)
        # get wavelength scale
        w0 = self.action.args.ccddata.header['CRVAL3']
        dw = self.action.args.ccddata.header['CD3_3']
        crpixw = self.action.args.ccddata.header['CRPIX3']
        # get all good wavelength range
        wgoo0 = self.action.args.ccddata.header['WAVGOOD0']
        if wgoo0 < 3650:
            wgoo0 = 3650.
        wgoo1 = self.action.args.ccddata.header['WAVGOOD1']
        # get all-inclusive wavelength range
        wall0 = self.action.args.ccddata.header['WAVALL0']
        wall1 = self.action.args.ccddata.header['WAVALL1']
        # get DAR padding in y
        pad_y = self.action.args.ccddata.header['DARPADY']
        # get sky subtraction status
        if 'SKYCOR' in self.action.args.ccddata.header:
            skycor = self.action.args.ccddata.header['SKYCOR']
        else:
            skycor = False
        # get telescope and atm. correction
        if 'TELESCOP' in self.action.args.ccddata.header:
            tel = self.action.args.ccddata.header['TELESCOP']
        else:
            tel = 'KeckI'
        if 'Keck' in tel:
            area = 760000.0
        else:
            area = -1.0
        # compute good y pixel ranges
        if w0 > 0. and dw > 0. and wgoo0 > 0. and wgoo1 > 0.:
            z0 = int((wgoo0 - w0) / dw) + 10
            z1 = int((wgoo1 - w0) / dw) - 10
        # wavelength scale
        w = w0 + z * dw
        # good spatial range
        gy0 = pad_y if pad_y > 1 else 1
        gy1 = sz[1] - (pad_y if pad_y > 2 else 2)
        # log results
        self.logger.info("Invsen. Pars: Y0, Y1, Z0, Z1, Wav0, Wav1: "
                         "%d, %d, %d, %d, %.3f, %.3f" %
                         (gy0, gy1, z0, z1, w[z0], w[z1]))
        # central wavelength
        cwv = self.action.args.cwave
        # find standard in slices
        # sum over wavelength
        tot = np.sum(self.action.args.ccddata.data[z0:z1, gy0:gy1, :], 0)
        # yy = np.arange(gy1-gy0) + gy0
        mxsl = -1.
        mxsg = 0.
        # for each slice
        for i in range(sz[2]):
            tstd = float(np.nanstd(tot[:, i]))
            if tstd > mxsg:
                mxsg = tstd
                mxsl = i
            # plot slice data, if requested
            if do_plots:
                py = tot[:, i]
                px = np.arange(len(py))
                p = figure(
                    title=self.action.args.stdlabel +
                    ' Std Slice %d (DAR padded)' % i,
                    x_axis_label="Position along slice",
                    y_axis_label="Flux summed over WLs",
                    plot_width=self.config.instrument.plot_width,
                    plot_height=self.config.instrument.plot_height)
                p.scatter(px, py, marker='x')
                bokeh_plot(p, self.context.bokeh_session)
                print("sl, std: %02d, %.3f" % (i, tstd))
                qstr = input("Next? <cr>, q - quit: ")
                if 'Q' in qstr.upper():
                    do_plots = False

        # relevant slices
        if self.action.args.ifunum == 1:    # Large slicer
            slset = 3
        elif self.action.args.ifunum == 2:  # Medium slicer
            slset = 5
        elif self.action.args.ifunum == 3:  # Small slicer
            slset = 12
        else:
            slset = 3
            self.logger.error("Assuming Large slicer: id undefined! - %d" %
                              self.action.args.ifunum)
        # Get set of slices needed for calculation
        sl0 = (mxsl - slset) if mxsl >= slset else 0
        sl1 = (mxsl + slset) if (mxsl + slset) <= sz[2]-1 else sz[2]-1
        # get y position of std
        cy, _ = find_peaks(tot[:, mxsl], height=np.nanmean(tot[:, mxsl]))
        cy = int(cy[0]) + gy0
        # log results
        self.logger.info("Std slices (DAR padded): "
                         "max, sl0, sl1, spatial cntrd: "
                         "%d, %d, %d, %.2f" % (mxsl, sl0, sl1, cy))
        # get dwave spectrum
        ofn = self.action.args.name
        delfn = strip_fname(ofn) + '_dcubed.fits'
        full_path = os.path.join(
            self.config.instrument.cwd,
            self.config.instrument.output_directory, delfn)
        if os.path.exists(full_path):
            dew = kcwi_fits_reader(full_path)[0]
            dwspec = dew.data[:, cy, mxsl]
            zeros = np.where(dwspec == 0)
            if len(zeros) > 0:
                dwspec[zeros] = dw
        else:
            dwspec = np.zeros(sz[0]) + dw
        # copy of input cube
        scub = self.action.args.ccddata.data.copy()
        # sky window width in pixels
        skywin = int(self.config.instrument.psfwid / self.action.args.xbinsize)
        # do sky subtraction, if needed
        if not skycor:
            self.logger.warning("Sky should have been subtraced already")
        # apply extinction correction
        ucub = scub.copy()  # uncorrected cube
        kcwi_correct_extin(scub, self.action.args.ccddata.header,
                           logger=self.logger)

        # get slice spectra y limits
        sy0 = (cy - skywin) if (cy - skywin) > 0 else 0
        sy1 = (cy + skywin) if (cy + skywin) < (sz[1]-1) else (sz[1]-1)
        # sum over y range
        slspec = np.sum(scub[:, sy0:sy1, :], 1)
        ulspec = np.sum(ucub[:, sy0:sy1, :], 1)
        # sum over slices
        obsspec = np.sum(slspec[:, sl0:sl1], 1)
        ubsspec = np.sum(ulspec[:, sl0:sl1], 1)
        # convert to e-/second
        obsspec /= expt
        ubsspec /= expt
        # check for zeros
        zeros = np.where(obsspec == 0.)
        if len(zeros) > 0:
            obsmean = np.nanmean(obsspec)
            obsspec[zeros] = obsmean
        # read in standard star spectrum
        stdfile = self.action.args.stdfile
        hdul = pf.open(stdfile)
        swl = hdul[1].data['WAVELENGTH']
        sflx = hdul[1].data['FLUX']
        sfw = hdul[1].data['FWHM']
        hdul.close()
        # get region of interest
        sroi = [i for i, v in enumerate(swl) if w[0] <= v <= w[-1]]
        nsroi = len(sroi)
        if nsroi <= 0:
            self.logger.error("No standard wavelengths in common")
            return self.action.args
        # expand range after checking for edges
        if sroi[0] > 0:
            sroi.insert(0, sroi[0]-1)
            nsroi += 1
        if sroi[-1] < (len(swl)-1):
            sroi.append(sroi[-1]+1)
            nsroi += 1
        # how many points?
        self.logger.info("Number of standard points = %d" % nsroi)
        # very sparsely sampled w.r.t. object
        if nsroi <= 5:
            self.logger.error("Not enough standard points")
            return self.action.args

        swl = swl[sroi]
        sflx = sflx[sroi]
        sfw = sfw[sroi]
        fwhm = np.max(sfw)
        self.logger.info("Reference spectrum FWHM used = %.1f (A)" % fwhm)
        # resample standard onto our wavelength grid
        rsint = interp1d(swl, sflx, kind='cubic', fill_value='extrapolate')
        rsflx = rsint(w)
        # get effective inverse sensitivity
        invsen = rsflx / obsspec
        # convert to photons/s/cm^2/(wl bin = dw)
        rspho = 5.03411250e7 * rsflx * w * dw
        # get effective area
        earea = ubsspec / rspho
        # correct to native bins
        earea *= dw/dwspec
        # Line masks
        lmasks = []
        # Header for line mask file
        lmhdr = []
        # default values (for BM)
        ford = 9        # fit order
        if 'BL' in self.action.args.grating:
            ford = 7
        elif 'BH' in self.action.args.grating:
            ford = 9
        # Adjust for dichroic fraction
        try:
            dichroic_fraction = self.action.args.ccddata.header['DICHFRAC']
        except KeyError:
            dichroic_fraction = 1.
        ford = int(ford * dichroic_fraction)
        if ford < 3:
            ford = 3
        self.logger.info("Fitting Invsens with polynomial order %d" % ford)
        # fit inverse sensitivity and effective area
        # get initial region of interest
        wl_good = [i for i, v in enumerate(w) if wgoo0 <= v <= wgoo1]
        nwl_good = len(wl_good)
        if nwl_good <= 0:
            self.logger.error("No good wavelengths to fit")
            return self.action.args
        wlm0 = wgoo0
        wlm1 = wgoo1
        # interactively set wavelength limits
        if self.config.instrument.plot_level >= 1:
            print("CHECKING WAVELENGTH LIMITS")
            print("Current WL limits: %.1f - %.1f Angstroms "
                  "(blue vertical lines)" % (wlm0, wlm1))
            print("A <cr> will accept current values, "
                  "or enter new values to avoid edge problems (if present).")
            print("Hover the cursor over spectrum to find wavelengths.")
            # yran = [np.min(obsspec), np.max(obsspec)]
            yran = [np.min(obsspec[wl_good]), np.max(obsspec[wl_good])]
            source = ColumnDataSource(data=dict(x=w, y=obsspec))
            done = False
            while not done:
                p = figure(
                    tooltips=[("x", "@x{0,0.0}"), ("y", "@y{0,0.0}")],
                    title=self.action.args.stdlabel + ' Obs Spec',
                    x_axis_label='Wave (A)',
                    y_axis_label='Intensity (e-)',
                    plot_width=self.config.instrument.plot_width,
                    plot_height=self.config.instrument.plot_height)
                p.line('x', 'y', line_color='black', source=source)
                p.line([wgoo0, wgoo0], yran, line_color='green',
                       legend_label='WAVGOOD')
                p.line([wgoo1, wgoo1], yran, line_color='green')
                p.line([wlm0, wlm0], yran, line_color='blue',
                       legend_label='LIMITS')
                p.line([wlm1, wlm1], yran, line_color='blue')
                p.line([cwv, cwv], yran, line_color='red', legend_label='CWAV')
                set_plot_lims(p, xlim=[wall0, wall1], ylim=yran)
                bokeh_plot(p, self.context.bokeh_session)
                qstr = input("New values? <float> <float> (or <cr> - done): ")
                if len(qstr) <= 0:
                    done = True
                else:
                    try:
                        wlm0 = float(qstr.split()[0])
                        wlm1 = float(qstr.split()[1])
                        if wlm1 < wlm0 or wlm0 < wall0 or wlm1 > wall1:
                            wlm0 = wgoo0
                            wlm1 = wgoo1
                            print("range/order error, try again")
                    except (IndexError, ValueError):
                        wlm0 = wgoo0
                        wlm1 = wgoo1
                        print("format error, try again")
            # update region of interest
            wl_good = [i for i, v in enumerate(w) if wlm0 <= v <= wlm1]
            nwl_good = len(wl_good)
        # END: interactively set wavelength limits

        # Look for mask data file
        # first in local directory
        local_lmfile = os.path.basename(stdfile).split('.fits')[0] + '.lmsk'
        if not os.path.exists(local_lmfile):
            # then in stds directory
            lmfile = stdfile.split('.fits')[0] + '.lmsk'
            if not os.path.exists(lmfile):
                lmfile = None
        else:
            lmfile = local_lmfile
        # if we found a mask file, read it in
        if lmfile is None:
            self.logger.info("No line mask file found")
        else:
            self.logger.info("Using line mask file %s" % lmfile)
            with open(lmfile) as lmf:
                lmasks_str = lmf.readlines()
            # parse file into mask list
            for lmws in lmasks_str:
                lmws = lmws.strip()
                # Collect header lines
                if lmws.startswith('#'):
                    lmhdr.append(lmws)
                    continue
                # Skip blank lines
                if len(lmws.strip()) < 1:
                    continue
                try:
                    data = lmws.split('#')[0]
                    # Parse comment on line
                    if '#' in lmws:
                        comm = lmws.split('#')[1].lstrip()
                    else:
                        comm = ''
                    # Parse line mask range
                    lm = [float(v) for v in data.split()]
                except ValueError:
                    print("bad line: %s" % lmws)
                    continue
                # Only collect good lines
                if len(lm) == 2 and lm[0] < lm[1]:
                    lmdict = {'w0': lm[0], 'w1': lm[1], 'com': comm}
                    lmasks.append(lmdict)
                else:
                    print("bad line: %s" % lmws)
        # Now interactively identify lines if requested
        if self.config.instrument.plot_level >= 1:
            yran = [np.min(obsspec[wl_good]), np.max(obsspec[wl_good])]
            # source = ColumnDataSource(data=dict(x=w, y=obsspec))
            print("MASKING SHARP FEATURES: ABSORPTION LINES/COSMIC RAYS")
            print("To mask, enter starting and stopping wavelengths "
                  "followed by an optional comment (string) for each line "
                  "you want to mask.")
            print("Current masks are shown as vertical dashed yellow lines.")
            print("Hover the cursor over spectrum to find wavelengths.")
            print("Mask values and comments are saved to this file: %s." %
                  local_lmfile)
            print("A <cr> with no line limits will accept current masks.")
            done = False
            while not done:
                p = figure(
                    tooltips=[("x", "@x{0.0}"), ("y", "@y{0.0}")],
                    title=self.action.args.stdlabel + ' Obs Spec',
                    x_axis_label='Wave (A)',
                    y_axis_label='Intensity (e-)',
                    plot_width=self.config.instrument.plot_width,
                    plot_height=self.config.instrument.plot_height)
                p.line(w, obsspec, line_color='black')
                p.line([wgoo0, wgoo0], yran, line_color='green',
                       legend_label='WAVGOOD')
                p.line([wgoo1, wgoo1], yran, line_color='green')
                p.line([wlm0, wlm0], yran, line_color='blue',
                       legend_label='LIMITS')
                p.line([wlm1, wlm1], yran, line_color='blue')
                p.line([cwv, cwv], yran, line_color='red', legend_label='CWAV')
                for ml in lmasks:
                    if wall0 < ml['w0'] < wall1 and wall0 < ml['w1'] < wall1:
                        p.line([ml['w0'], ml['w0']], yran, line_color='orange',
                               line_dash='dashed')
                        p.line([ml['w1'], ml['w1']], yran, line_color='orange',
                               line_dash='dashed')
                set_plot_lims(p, xlim=[wall0, wall1], ylim=yran)
                bokeh_plot(p, self.context.bokeh_session)
                qstr = input("Mask line: wavelength start stop (Ang) comment "
                             "(str) <float> <float> <str> (<cr> - done): ")
                if len(qstr) <= 0:
                    done = True
                else:
                    try:
                        newl = {'w0': float(qstr.split()[0]),
                                'w1': float(qstr.split()[1]),
                                'com': " ".join(qstr.split()[2:]).strip()}
                        # newl = [float(val) for val in qstr.split()]
                    except ValueError:
                        print("bad line: %s" % qstr)
                        continue
                    if wlm0 < newl['w0'] < wlm1 and wlm0 < newl['w1'] < wlm1:
                        lmasks.append(newl)
                    else:
                        print("line mask outside wl range: %s" % qstr)
        # END: interactively identify lines
        # Write out a local copy of line masks so user can edit
        if len(lmasks) > 0:
            local_lmfile = os.path.basename(
                stdfile).split('.fits')[0] + '.lmsk'
            with open(local_lmfile, 'w') as lmf:
                for lmh in lmhdr:
                    lmf.write(lmh)
                for lm in lmasks:
                    if len(lm['com']) == 0:
                        lmf.write("%.2f %.2f\n" % (lm['w0'], lm['w1']))
                    else:
                        lmf.write("%.2f %.2f # %s\n" % (lm['w0'], lm['w1'],
                                                        lm['com']))
        # set up fitting vectors, flux, waves, measure errors
        sf = invsen[wl_good]   # dependent variable
        af = earea[wl_good]    # effective area
        wf = w[wl_good]        # independent variable
        bf = obsspec[wl_good]  # input electrons
        rf = rsflx[wl_good]    # reference flux
        mw = np.ones(nwl_good, dtype=float)     # weights
        use = np.ones(nwl_good, dtype=int)      # toggles for usage
        # loop over line masks and apply them
        for ml in lmasks:
            roi = [i for i, v in enumerate(wf) if ml['w0'] <= v <= ml['w1']]
            nroi = len(roi)
            if nroi > 0:
                use[roi] = 0
                self.logger.info("Masking line at %.2f to %.2f (A)"
                                 % (ml['w0'], ml['w1']))
        # ignore bad points by setting large errors
        mf = []
        ef = []
        ww = []
        used = []
        not_used = []
        for i in range(len(use)):
            if use[i] == 1:
                used.append(i)
            else:
                mw[i] = 1.e-9
                mf.append(sf[i])
                ef.append(100.*af[i]/area)
                ww.append(wf[i])
                not_used.append(i)

        # initial polynomial fit of inverse sensitivity
        wf0 = np.min(wf)
        res = np.polyfit(wf-wf0, sf, deg=ford, w=mw)
        finvsen = np.polyval(res, w-wf0)
        sinvsen = np.polyval(res, wf-wf0)
        calspec = obsspec * finvsen
        scalspec = bf * sinvsen
        # initial polynomial fit of effective area
        res = np.polyfit(wf-wf0, af, ford, w=mw)
        fearea = np.polyval(res, w-wf0)
        # calculate residuals
        resid = 100.0 * (scalspec - rf) / rf
        if len(not_used) > 0:
            rbad = resid[not_used]
        else:
            rbad = None
        rsd_mean = float(np.nanmean(resid[used]))
        rsd_stdv = float(np.nanstd(resid[used]))
        self.logger.info("Calibration residuals = %f +- %f %%" %
                         (rsd_mean, rsd_stdv))
        # plots
        peff = None
        pivs = None
        pcal = None
        prsd = None
        # interactively adjust fit
        if self.config.instrument.plot_level >= 1:
            done = False
            while not done:
                yran = [np.min(100.*af/area), np.max(100.*af/area)]
                effmax = np.nanmax(100.*fearea/area)
                effmean = np.nanmean(100.*fearea/area)
                peff = figure(
                    title=self.action.args.stdlabel + ' Efficiency',
                    x_axis_label='Wave (A)',
                    y_axis_label='Effective Efficiency (%)',
                    plot_width=self.config.instrument.plot_width,
                    plot_height=self.config.instrument.plot_height)
                peff.line(wf, 100.*af/area, line_color='black',
                          legend_label='Data')
                peff.line(w, 100.*fearea/area, line_color='red',
                          legend_label='Fit')
                peff.scatter(ww, ef, marker='x', legend_label='Rej')
                peff.line([wlm0, wlm0], yran, line_color='green',
                          legend_label='WL lims')
                peff.line([wlm1, wlm1], yran, line_color='green')
                peff.line([wall0, wall1], [effmax, effmax],
                          line_color='black', line_dash='dashed')
                peff.line([wall0, wall1], [effmean, effmean],
                          line_color='black', line_dash='dashdot')
                set_plot_lims(peff, xlim=[wall0, wall1], ylim=yran)
                bokeh_plot(peff, self.context.bokeh_session)
                if self.config.instrument.plot_level >= 1:
                    input("Next? <cr>: ")
                else:
                    time.sleep(2. * self.config.instrument.plot_pause)

                yran = [np.min(sf), np.max(sf)]
                pivs = figure(
                    title=self.action.args.stdlabel + ' Inverse sensitivity',
                    x_axis_label='Wave (A)',
                    y_axis_label='Invserse Sensitivity (Flux/e-/s)',
                    y_axis_type='log',
                    plot_width=self.config.instrument.plot_width,
                    plot_height=self.config.instrument.plot_height)
                pivs.line(wf, sf, line_color='black', legend_label='Data')
                pivs.line(w, finvsen, line_color='red', legend_label='Fit')
                pivs.scatter(ww, mf, marker='x', legend_label='Rej')
                pivs.line([wlm0, wlm0], yran, line_color='green',
                          legend_label='WL lims')
                pivs.line([wlm1, wlm1], yran, line_color='green')
                set_plot_lims(pivs, xlim=[wall0, wall1], ylim=yran)
                bokeh_plot(pivs, self.context.bokeh_session)
                if self.config.instrument.plot_level >= 1:
                    input("Next? <cr>: ")
                else:
                    time.sleep(2. * self.config.instrument.plot_pause)

                yran = [np.min(calspec[wl_good]), np.max(calspec[wl_good])]
                pcal = figure(
                    title=self.action.args.stdlabel + ' Calibrated',
                    x_axis_label='Wave (A)',
                    y_axis_label='Flux (ergs/s/cm^2/A)',
                    plot_width=self.config.instrument.plot_width,
                    plot_height=self.config.instrument.plot_height)
                pcal.line(w, calspec, line_color='black', legend_label='Obs')
                pcal.line(w, rsflx, line_color='red', legend_label='Ref')
                pcal.line([wlm0, wlm0], yran, line_color='green',
                          legend_label='WL lims')
                pcal.line([wlm1, wlm1], yran, line_color='green')
                set_plot_lims(pcal, xlim=[wall0, wall1], ylim=yran)
                bokeh_plot(pcal, self.context.bokeh_session)
                if self.config.instrument.plot_level >= 1:
                    input("Next? <cr>: ")
                else:
                    time.sleep(2. * self.config.instrument.plot_pause)

                yran = [np.min(resid), np.max(resid)]
                prsd = figure(
                    title=self.action.args.stdlabel +
                    ' Residuals = %.1f +- %.1f (%%)' % (rsd_mean, rsd_stdv),
                    x_axis_label='Wave (A)',
                    y_axis_label='Obs - Ref / Ref (%)',
                    plot_width=self.config.instrument.plot_width,
                    plot_height=self.config.instrument.plot_height)
                prsd.line(wf, resid, line_color='black',
                          legend_label='Obs - Ref (%)')
                if len(not_used) > 0:
                    prsd.scatter(ww, rbad, marker='x', legend_label='Rej')
                prsd.line([wlm0, wlm0], yran, line_color='green',
                          legend_label='WL lims')
                prsd.line([wlm1, wlm1], yran, line_color='green')
                prsd.line([wall0, wall1], [rsd_mean, rsd_mean],
                          line_color='red')
                prsd.line([wall0, wall1],
                          [rsd_mean+rsd_stdv, rsd_mean+rsd_stdv],
                          line_color='black', line_dash='dashed')
                prsd.line([wall0, wall1],
                          [rsd_mean-rsd_stdv, rsd_mean-rsd_stdv],
                          line_color='black', line_dash='dashed')
                set_plot_lims(prsd, xlim=[wall0, wall1], ylim=yran)
                bokeh_plot(prsd, self.context.bokeh_session)
                if self.config.instrument.plot_level >= 1:
                    qstr = input("Current fit order = %d, "
                                 "New fit order? <int>, <cr> - done: " % ford)
                    if len(qstr) <= 0:
                        done = True
                    else:
                        try:
                            ford = int(qstr)
                            # update fit of inverse sensitivity
                            res = np.polyfit(wf - wf0, sf, deg=ford, w=mw)
                            finvsen = np.polyval(res, w - wf0)
                            sinvsen = np.polyval(res, wf - wf0)
                            calspec = obsspec * finvsen
                            scalspec = bf * sinvsen
                            # update polynomial fit of effective area
                            res = np.polyfit(wf - wf0, af, ford, w=mw)
                            fearea = np.polyval(res, w - wf0)
                            # re-calculate residuals
                            resid = 100.0 * (scalspec - rf) / rf
                            if len(not_used) > 0:
                                rbad = resid[not_used]
                            else:
                                rbad = None
                            rsd_mean = float(np.nanmean(resid[used]))
                            rsd_stdv = float(np.nanstd(resid[used]))
                            self.logger.info(
                                "Calibration residuals = %f +- %f %%" %
                                (rsd_mean, rsd_stdv))
                        except ValueError:
                            print("Bad fit order, try again")
                else:
                    done = True
                    time.sleep(2. * self.config.instrument.plot_pause)
            # log results
            effmax = float(np.nanmax(100. * fearea / area))
            effmean = float(np.nanmean(100. * fearea / area))
            self.logger.info("Peak, mean efficiency (%%): %.1f, %.1f" %
                             (effmax, effmean))
            self.logger.info("Fit order = %d" % ford)
            # output plots
            pfname = "std_%05d_%s_%s_%s_%d" % (
                self.action.args.ccddata.header['FRAMENO'],
                stdname, self.action.args.grating.strip(),
                self.action.args.ifuname.strip(), int(self.action.args.cwave))
            # Save plots
            save_plot(peff, filename=pfname + '_eff.png')      # Efficiency
            save_plot(pivs, filename=pfname + '_invsens.png')  # Inv. Sens.
            save_plot(prsd, filename=pfname + '_resid.png')    # Residuals
            save_plot(pcal, filename=pfname + '_cal.png')      # Calibrated

        log_string = MakeInvsens.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        # write out effective inverse sensitivity

        # update inverse sensitivity header
        hdr = self.action.args.ccddata.header.copy()
        hdr['HISTORY'] = log_string
        hdr['INVSENS'] = (True, 'effective inv. sens. spectrum?')
        hdr['INVSW0'] = (w[z0], 'low wavelength for eff. inv. sens.')
        hdr['INVSW1'] = (w[z1], 'high wavelength for eff. inv. sens.')
        hdr['INVSZ0'] = (z0, 'low wave pixel for eff. inv. sens.')
        hdr['INVSZ1'] = (z1, 'high wave pixel for eff. inv. sens.')
        hdr['INVSY0'] = (gy0, 'low spatial pixel for eff. inv. sens.')
        hdr['INVSY1'] = (gy1, 'high spatial pixel for eff. inv. sens.')
        hdr['INVSLMX'] = (mxsl, 'brightest std star slice')
        hdr['INVSL0'] = (sl0, 'lowest std star slice summed')
        hdr['INVSL1'] = (sl1, 'highest std star slice summed')
        hdr['INVSLY'] = (cy, 'spatial pixel position of std within slice')
        hdr['INVFW0'] = (wlm0, 'low wavelength for fits')
        hdr['INVFW1'] = (wlm1, 'high wavelength for fits')
        hdr['INVFORD'] = (ford, 'fit order')
        hdr['EXPTIME'] = (1., 'effective exposure time (seconds)')
        hdr['XPOSURE'] = (1., 'effective exposure time (seconds)')

        # remove old WCS
        del hdr['RADESYS']
        del hdr['EQUINOX']
        del hdr['LONPOLE']
        del hdr['LATPOLE']
        del hdr['NAXIS2']
        if 'NAXIS3' in hdr:
            del hdr['NAXIS3']
        del hdr['CTYPE1']
        del hdr['CTYPE2']
        del hdr['CTYPE3']
        del hdr['CUNIT1']
        del hdr['CUNIT2']
        del hdr['CUNIT3']
        del hdr['CNAME1']
        del hdr['CNAME2']
        del hdr['CNAME3']
        del hdr['CRVAL1']
        del hdr['CRVAL2']
        del hdr['CRVAL3']
        del hdr['CRPIX1']
        del hdr['CRPIX2']
        del hdr['CRPIX3']
        del hdr['CD1_1']
        del hdr['CD1_2']
        del hdr['CD2_1']
        del hdr['CD2_2']
        del hdr['CD3_3']

        # set wavelength axis WCS values
        hdr['WCSDIM'] = 1
        hdr['CTYPE1'] = ('AWAV', 'Air Wavelengths')
        hdr['CUNIT1'] = ('Angstrom', 'Wavelength units')
        hdr['CNAME1'] = ('KCWI INVSENS Wavelength', 'Wavelength name')
        hdr['CRVAL1'] = (w0, 'Wavelength zeropoint')
        hdr['CRPIX1'] = (crpixw, 'Wavelength reference pixel')
        hdr['CDELT1'] = (dw, 'Wavelength Angstroms per pixel')

        ofn = self.action.args.name
        invsname = strip_fname(ofn) + '_' + suffix + '.fits'
        eaname = strip_fname(ofn) + '_ea.fits'

        # set units
        invsens_u = u.erg / (u.angstrom * u.cm ** 2 * u.s * u.electron)
        # output inverse sensitivity
        out_invsens = CCDData(np.asarray([invsen, finvsen, obsspec]),
                              meta=hdr, unit=invsens_u)
        kcwi_fits_writer(out_invsens, output_file=invsname,
                         output_dir=self.config.instrument.output_directory)
        self.context.proctab.update_proctab(frame=out_invsens, suffix=suffix,
                                            newtype='INVSENS',
                                            filename=self.action.args.name)
        # output effective area
        ea_u = u.cm ** 2 / u.angstrom
        out_ea = CCDData(np.asarray([earea, fearea]), meta=hdr, unit=ea_u)
        kcwi_fits_writer(out_ea, output_file=eaname,
                         output_dir=self.config.instrument.output_directory)
        self.context.proctab.update_proctab(frame=out_ea, suffix='ea',
                                            newtype='EAREA',
                                            filename=self.action.args.name)

        return self.action.args
    # END: class MakeInvsens()



from keckdrpframework.primitives.base_img import BaseImg
from kcwidrp.primitives.kcwi_file_primitives import kcwi_fits_reader, \
    kcwi_fits_writer, strip_fname

import os
import ccdproc
import numpy as np
from astropy.stats import mad_std


def inv_median(a):
    return 1.0/np.median(a)


class StackFlats(BaseImg):
    """
    Stack flat images

    Generate a stacked flat frame based on one of the following instrument
    config parameters:

        * flat_min_nframes: defaults to 6
        * dome_min_nframes: defaults to 3
        * twiflat_min_nframes: defaults to 1

    It is assumed that each frame is well-exposed and the combine
    method 'average' will be used.  A high sigma clipping of 2.0 is used to
    help with the CRs.

    Uses the ccdproc.combine routine to peform the stacking.

    Writes out the following files and adds entries in the proc table depending
    on the input IMTYPE:

        * FLATLAMP - writes out \*_sflat.fits and puts SFLAT in proc table
        * DOMEFLAT - writes out \*_sdome.fits and puts SDOME in proc table
        * TWIFLAT -  writes out \*_stwif.fits and puts STWIF in proc table

    """

    def __init__(self, action, context):
        BaseImg.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        """
        Checks if we can build a stacked frame based on the processing table
        """
        # get list of input flats
        self.logger.info("Checking precondition for StackFlats")
        self.stacked_list = self.context.proctab.search_proctab(
            frame=self.action.args.ccddata,
            target_type=self.action.args.stack_type,
            target_group=self.action.args.groupid)
        if len(self.stacked_list) > 0:
            self.logger.info(f"already have {len(self.stacked_list)},"
                             f" stacked flats, expecting 0")
            return False
        else:
            self.combine_list = self.context.proctab.search_proctab(
                frame=self.action.args.ccddata,
                target_type=self.action.args.want_type,
                target_group=self.action.args.groupid)
            self.logger.info(f"pre condition got {len(self.combine_list)},"
                             f" expecting {self.action.args.min_files}")
            # do we meet the criterion?
            if len(self.combine_list) >= self.action.args.min_files:
                return True
            else:
                self.action.new_event = None
                return False

    def _perform(self):
        """
        Returns an Argument() with the parameters that depends on this operation
        """
        method = 'average'
        sig_up = 2.0
        suffix = self.action.args.stack_type.lower()

        self.logger.info("Stacking flats using method %s" % method)

        combine_list = list(self.combine_list['filename'])
        # get flat stack output name
        stname = strip_fname(combine_list[0]) + '_' + suffix + '.fits'
        stack = []
        stackf = []
        fmeds = []
        for flat in combine_list:
            # get flat intensity (int) image file name in redux directory
            stackf.append(strip_fname(flat) + '_intd.fits')
            flatfn = os.path.join(self.config.instrument.cwd,
                                  self.config.instrument.output_directory,
                                  stackf[-1])
            # using [0] gets just the image data
            f = kcwi_fits_reader(flatfn)[0]
            # Set mask to None to prevent ccdproc.combine from masking
            f.mask = None
            # get scale
            fmed = float(np.median(f))
            self.logger.info("%s - median: %.2f" % (flat, fmed))
            fmeds.append(fmed)
            stack.append(f)

        stacked = ccdproc.combine(stack, method=method, sigma_clip=True,
                                  scale=inv_median,
                                  sigma_clip_low_thresh=None,
                                  sigma_clip_high_thresh=sig_up,
                                  sigma_clip_func=np.ma.median,
                                  sigma_clip_dev_func=mad_std)

        # re-scale stacked image
        stacked.data = stacked.data * np.median(fmeds)

        # Get the bad pixel mask out of one of the flats (is the same for all)
        # and add it to the stacked flat as the stack's mask
        last_flat_name = strip_fname(combine_list[-1]) + '_intd.fits'
        last_flat_path = os.path.join(self.config.instrument.cwd,
                                      self.config.instrument.output_directory,
                                      last_flat_name)
        last_flat = kcwi_fits_reader(last_flat_path)[0]
        stacked.mask = last_flat.mask
        
        stacked.header['IMTYPE'] = self.action.args.stack_type
        stacked.header['NSTACK'] = (len(combine_list),
                                    'number of images stacked')
        stacked.header['STCKMETH'] = (method, 'method used for stacking')
        stacked.header['STCKSIGU'] = (sig_up,
                                      'Upper sigma rejection for stacking')
        for ii, fname in enumerate(stackf):
            stacked.header['STACKF%d' % (ii + 1)] = (fname, "stack input file")

        log_string = StackFlats.__module__
        stacked.header['HISTORY'] = log_string

        # output stacked flat
        kcwi_fits_writer(stacked, output_file=stname,
                         output_dir=self.config.instrument.output_directory)

        self.context.proctab.update_proctab(frame=stacked, suffix=suffix,
                                            newtype=self.action.args.stack_type,
                                            filename=stacked.header['OFNAME'])
        self.context.proctab.write_proctab(tfil=self.config.instrument.procfile)

        self.logger.info(log_string)
        return self.action.args

    # END: class StackFlats()



from keckdrpframework.primitives.base_primitive import BasePrimitive
from kcwidrp.primitives.kcwi_file_primitives import strip_fname
import os
import numpy as np
from kcwidrp.core import geometric as tf
import pickle


class SolveGeom(BasePrimitive):
    """
    Solve the overall geometry of the IFU.

    Takes individual bar arc spectra and generates a fit for each slice, which
    contains five bar arc spectra each.  Given that there are only five points
    in the 'x' or spatial direction and many more in the 'y' or wavelength
    direction, an asymmetric polynomial is fit with order 2 in the x and order
    4 in the y directions.  The wavelength coverage of the observation is
    recorded in parameters with the range that includes all data being in
    the waveall0 and waveall1 parameters, and the range that includes only
    good data in wavegood0 and wavegood1 parameters.  The middle of the
    wavelength range is recorded in the wavemid parameter.

    Forward and inverse transforms, along with all the parameters for the
    geometric fit are written out as a python pickled dictionary in a
    \*_geom.pkl file.

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.action.args.geometry_file = None
        self.action.args.x0out = None
        self.action.args.wave0out = None
        self.action.args.wave1out = None
        self.action.args.wavegood0 = None
        self.action.args.wavegood1 = None
        self.action.args.waveall0 = None
        self.action.args.waveall1 = None
        self.action.args.wavemid = None
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        self.logger.info("Checking for master arc")
        if 'MARC' in self.action.args.ccddata.header['IMTYPE']:
            return True
        else:
            return False

    def _perform(self):
        self.logger.info("Solving overall geometry")

        # Get some geometry constraints
        goody0 = 0
        goody1 = max(self.action.args.xsvals)
        # N&S limits
        goodnsy0 = self.action.args.shufrows
        goodnsy1 = goodnsy0 + self.action.args.shufrows
        # Calculate wavelength ranges
        y0wvs = []
        y1wvs = []
        # N&S ranges
        y0nswvs = []
        y1nswvs = []
        # Get wavelength extremes for each bar
        for fcfs in self.action.args.fincoeff:
            y0wvs.append(float(np.polyval(fcfs, goody0)))
            y1wvs.append(float(np.polyval(fcfs, goody1)))
            y0nswvs.append(float(np.polyval(fcfs, goodnsy0)))
            y1nswvs.append(float(np.polyval(fcfs, goodnsy1)))
        # Now get ensemble extremes
        y0max = max(y0wvs)
        y0min = min(y0wvs)
        y1max = max(y1wvs)
        y1min = min(y1wvs)
        y0nsmax = max(y0nswvs)
        y0nsmin = min(y0nswvs)
        y1nsmax = max(y1nswvs)
        y1nsmin = min(y1nswvs)
        # Cube trimming wavelengths
        trimw0 = y0min
        trimw1 = y1max
        # Check for negative dispersion
        if trimw0 > trimw1:
            trimw0 = y1min
            trimw1 = y0max
        # Calculate output wavelengths
        dwout = self.action.args.dwout
        ndels = int((trimw0 - self.config.instrument.WAVEFID) / dwout)
        self.action.args.wave0out = \
            self.config.instrument.WAVEFID + float(ndels) * dwout
        ndels = int((trimw1 - self.config.instrument.WAVEFID) / dwout)
        self.action.args.wave1out = \
            self.config.instrument.WAVEFID + float(ndels) * dwout
        self.logger.info("WAVE RANGE: %.2f - %.2f" %
                         (self.action.args.wave0out, self.action.args.wave1out))
        # Calculate wavelength limits
        self.action.args.wavegood0 = min([y0max, y1max])
        self.action.args.wavegood1 = max([y0min, y1min])
        self.action.args.waveall0 = min([y0min, y1min])
        self.action.args.waveall1 = max([y0max, y1max])
        self.action.args.wavemid = np.average([self.action.args.wavegood0,
                                               self.action.args.wavegood1,
                                               self.action.args.waveall0,
                                               self.action.args.waveall1])
        self.action.args.wavensgood0 = min([y0nsmax, y1nsmax])
        self.action.args.wavensgood1 = max([y0nsmin, y1nsmin])
        self.action.args.wavensall0 = min([y0nsmin, y1nsmin])
        self.action.args.wavensall1 = max([y0nsmax, y1nsmax])
        self.action.args.wavensmid = np.average([self.action.args.wavensgood0,
                                                 self.action.args.wavensgood1,
                                                 self.action.args.wavensall0,
                                                 self.action.args.wavensall1])
        self.logger.info("WAVE  GOOD: %.2f - %.2f" %
                         (self.action.args.wavegood0,
                          self.action.args.wavegood1))
        self.logger.info("WAVE   ALL: %.2f - %.2f" %
                         (self.action.args.waveall0, self.action.args.waveall1))
        self.logger.info("WAVE   MID: %.2f" % self.action.args.wavemid)
        # Start setting up slice transforms
        self.action.args.x0out = \
            int(self.action.args.reference_bar_separation / 2.) + 1
        self.refoutx = np.arange(0, 5) * \
            self.action.args.reference_bar_separation + self.action.args.x0out
        # Variables for output control points
        srcw = []
        # Loop over source control points
        for ixy, xy in enumerate(self.action.args.source_control_points):
            # Calculate y wavelength
            yw = float(np.polyval(
                self.action.args.fincoeff[self.action.args.bar_id[ixy]], xy[1]))
            # Convert to output pixels
            yw = (yw - self.action.args.wave0out) / dwout
            srcw.append([xy[0], yw])
        # Use extremes to define output size
        ysize = int((self.action.args.waveall1 - self.action.args.wave0out)
                    / dwout)
        xsize = int(5. * self.action.args.reference_bar_separation) + 1
        self.logger.info("Output slices will be %d x %d px" % (xsize, ysize))
        # Now loop over slices and get relevant control points for each slice
        # Output variables
        xl0_out = []
        xl1_out = []
        tform_list = []
        invtf_list = []
        # Loop over 24 slices
        for isl in range(0, 24):
            # Get control points
            xw = []
            yw = []
            xi = []
            yi = []
            # Loop over all control points
            for ixy, xy in enumerate(srcw):
                # Only use the ones for this slice
                if self.action.args.slice_id[ixy] == isl:
                    # Index in to reference output x array
                    ib = self.action.args.bar_id[ixy] % 5
                    # Geometrically corrected control points
                    xw.append(self.refoutx[ib])
                    yw.append(xy[1])
                    # Input control points
                    xi.append(self.action.args.destination_control_points[
                                  ixy][0])
                    yi.append(self.action.args.destination_control_points[
                                  ixy][1])
            # get image limits
            xl0 = int(min(xi) - self.action.args.reference_bar_separation)
            if xl0 < 0:
                xl0 = 0
            xl1 = int(max(xi) + self.action.args.reference_bar_separation)
            if xl1 > (self.action.args.ccddata.data.shape[0] - 1):
                xl1 = self.action.args.ccddata.data.shape[0] - 1
            # Store for output
            xl0_out.append(xl0)
            xl1_out.append(xl1)
            self.logger.info("Slice %d arc image x limits: %d - %d" %
                             (isl, xl0, xl1))
            # adjust control points
            xit = [x - float(xl0) for x in xi]
            # fit transform
            dst = np.column_stack((xit, yi))
            src = np.column_stack((xw, yw))
            self.logger.info("Fitting wavelength and spatial control points")
            tform = tf.estimate_transform('asympolynomial', src, dst,
                                          order=(2, 4))
            invtf = tf.estimate_transform('asympolynomial', dst, src,
                                          order=(2, 4))
            # Store for output
            tform_list.append(tform)
            invtf_list.append(invtf)
        # Pixel scales
        pxscl = self.config.instrument.PIXSCALE * self.action.args.xbinsize
        ifunum = self.action.args.ifunum
        if ifunum == 2:
            slscl = self.config.instrument.SLICESCALE / 2.0
        elif ifunum == 3:
            slscl = self.config.instrument.SLICESCALE / 4.0
        else:
            slscl = self.config.instrument.SLICESCALE
        # Dichroic fraction
        try:
            dichroic_fraction = self.action.args.dichroic_fraction
        except AttributeError:
            dichroic_fraction = 1.

        # Package geometry data
        ofname = self.action.args.ccddata.header['OFNAME']
        self.action.args.geometry_file = os.path.join(
            self.config.instrument.output_directory,
            strip_fname(ofname) + '_geom.pkl')
        if os.path.exists(self.action.args.geometry_file):
            self.logger.error("Geometry file already exists: %s" %
                              self.action.args.geometry_file)
        else:
            geom = {
                "geom_file": self.action.args.geometry_file,            # output save file for geom struct
                "xsize": xsize, "ysize": ysize,                         # xy size of image
                "pxscl": pxscl, "slscl": slscl,                         # degrees per unbinned spatial and slice pixel
                "cbarsno": self.action.args.contbar_image_number,       # cont bars image number
                "cbarsfl": self.action.args.contbar_image,              # cont bars image name
                "arcno": self.action.args.arc_number,                   # arc image number
                "arcfl": self.action.args.arc_image,                    # arc image name
                "barsep": self.action.args.reference_bar_separation,    # reference delta x between bars (pix)
                "bar0": self.action.args.x0out,                         # output spatial zeropoint (unbinned pix)
                "waveall0": self.action.args.waveall0,                  # low wavelength that includes all data
                "waveall1": self.action.args.waveall1,                  # high wavelength that includes all data
                "wavegood0": self.action.args.wavegood0,                # low wavelength that includes good data
                "wavegood1": self.action.args.wavegood1,                # high wavelength that includes good data
                "wavemid": self.action.args.wavemid,                    # wavelength of the middle of the range
                "wavensall0": self.action.args.wavensall0,              # low wavelength that includes all masked data
                "wavensall1": self.action.args.wavensall1,              # high wavelength that includes all masked data
                "wavensgood0": self.action.args.wavensgood0,            # low wavelength that includes good masked data
                "wavensgood1": self.action.args.wavensgood1,            # high wavelength that includes good masked data
                "wavensmid": self.action.args.wavensmid,                # wavelength of the middle of the masked range
                "dich_frac": dichroic_fraction,                         # fraction of wavelength range impacted by dichroic
                "dwout": dwout,                                         # output disperson (Ang/pix)
                "wave0out": self.action.args.wave0out,                  # output wavelength zeropoint
                "wave1out": self.action.args.wave1out,                  # output ending wavelength
                "avwvsig": self.action.args.av_bar_sig,                 # average bar wavelength sigma (Ang)
                "sdwvsig": self.action.args.st_bar_sig,                 # standard deviation of bar sigmas (Ang)
                "xl0": xl0_out, "xl1": xl1_out,                         # slice spatial limits
                "tform": tform_list, "invtf": invtf_list                # transform and inverse transforms for each slice
            }
            with open(self.action.args.geometry_file, 'wb') as ofile:
                pickle.dump(geom, ofile)
            self.logger.info("Geometry written to: %s" %
                             self.action.args.geometry_file)

        log_string = SolveGeom.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        return self.action.args
    # END: class SolveGeom()



from astropy.io import fits as pf
import glob
import numpy as np
from matplotlib import pyplot as pl

flist = glob.glob('*.fits')
print("Found %d fits files" % len(flist))

pl.ioff()

for f in flist:
    ff = pf.open(f)
    wave = ff[1].data['WAVELENGTH']
    flux = ff[1].data['FLUX']
    ff.close()
    name = f.split('.fits')[0]
    wlmin = np.nanmin(wave)
    wlmax = np.nanmax(wave)
    wllen = len(wave)
    print("%20s %10.2f %10.2f %d" % (f, wlmin, wlmax, wllen))
    fig = pl.figure()
    pl.yscale('log')
    pl.plot(wave, flux, "+")
    pl.xlim([wlmin-100., np.min([12000, wlmax])+100.])
    if wlmax > 12100:
        seen = flux[np.where(wave <= 12100.)]
        flmin = np.nanmin(seen)
        flmax = np.nanmax(seen)
        pl.ylim((flmin, flmax))
    pl.title("%s: %.2f - %.2f A, %d pts" % (name, wlmin, wlmax, wllen))
    pl.xlabel('WAVELENGTH')
    pl.ylabel('FLUX')
    pl.savefig(name+'.png')
    pl.close(fig)




from keckdrpframework.primitives.base_primitive import BasePrimitive


class ProcessDark(BasePrimitive):
    """
    Preliminary processing of dark images.
    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _perform(self):
        return self.action.args



"""
Asymmetric Polynomial Transform Code

Shamelessly pilfered from scikit image _geometric, which contains all the
transforms in the standard repertoir, but which did not include a 2D polynomial
transform that could have different orders for X and Y.

The asymmetric polynomial algorithm was ported to Python from the IDL routine,
mod_polywarp.pro, written by M. Matuszewski.  This routine was part of the
original IDL KCWI pipeline, which can be found at:

    * https://github.com/Keck-DataReductionPipelines/KcwiDRP

The resulting transform is compatible with the scikit image transform.warp
function.

"""
import math
import numpy as np
from scipy import linalg


class GeometricTransform(object):
    """Base class for geometric transformations.

    """
    def __call__(self, coords):
        """Apply forward transformation.

        Parameters
        ----------
        coords : (N, 2) array
            Source coordinates.

        Returns
        -------
        coords : (N, 2) array
            Destination coordinates.

        """
        raise NotImplementedError()

    def inverse(self, coords):
        """Apply inverse transformation.

        Parameters
        ----------
        coords : (N, 2) array
            Destination coordinates.

        Returns
        -------
        coords : (N, 2) array
            Source coordinates.

        """
        raise NotImplementedError()

    def residuals(self, src, dst):
        """Determine residuals of transformed destination coordinates.

        For each transformed source coordinate the euclidean distance to the
        respective destination coordinate is determined.

        Parameters
        ----------
        src : (N, 2) array
            Source coordinates.
        dst : (N, 2) array
            Destination coordinates.

        Returns
        -------
        residuals : (N, ) array
            Residual for coordinate.

        """
        return np.sqrt(np.sum((self(src) - dst)**2, axis=1))

    def __add__(self, other):
        """Combine this transformation with another.

        """
        raise NotImplementedError()


class AsymmetricPolynomialTransform(GeometricTransform):
    """2D polynomial transformation.

    Has the following form::

        X = sum[j=0:order]( sum[i=0:j]( a_ji * x**(j - i) * y**i ))
        Y = sum[j=0:order]( sum[i=0:j]( b_ji * x**(j - i) * y**i ))

    Parameters
    ----------
    params : (2, N) array, optional
        Polynomial coefficients where `N * 2 = (order + 1) * (order + 2)`. So,
        a_ji is defined in `params[0, :]` and b_ji in `params[1, :]`.

    Attributes
    ----------
    params : (2, N) array
        Polynomial coefficients where `N * 2 = (order + 1) * (order + 2)`. So,
        a_ji is defined in `params[0, :]` and b_ji in `params[1, :]`.

    """

    def __init__(self, params=None):
        if params is None:
            # default to transformation which preserves original coordinates
            params = np.array([[0, 1, 0], [0, 0, 1]])
        if params.shape[0] != 2:
            raise ValueError("invalid shape of transformation parameters")
        self.params = params

    def estimate(self, src, dst, order=(2, 2)):
        """Generate a pair of warping polynomial kernels in two
           dimensions (Kx, Ky) taking  starting coordinates (Xo,Yo) to
           ending coordinates (Xi,Yi). The polynomial degrees in the
           two dimensions are allowed to differ:

           Xi = Sum (i,0,Degree[1]) (j,0,Degree[0]) Kx[i,j] Xo^j Yo^i
           Yi = Sum (i,0,Degree[1]) (j,0,Degree[0]) Ky[i,j] Xo^j Yo^i

           The degrees need not be equal, but the number of points provided
           must be greater than or equal to

           number of points = (Degree[0]+1)*(Degree[1]+1)

        Parameters
        ----------
        src : (N, 2) array
            Source coordinates.
        dst : (N, 2) array
            Destination coordinates.
        order : (2) array, int
            Polynomial order (number of coefficients is order + 1).

        Returns
        -------
        success : bool
            True, if model estimation succeeds.

        """
        if isinstance(order, int):
            order = (order, order)

        m = order[0] + 1
        n = order[1] + 1
        efford = order[0] + order[1] + 1

        plen = int(efford * (efford + 1) / 2)

        xi = dst[:, 0]
        yi = dst[:, 1]
        xo = src[:, 0]
        yo = src[:, 1]
        npts = src.shape[0]

        W = np.zeros((npts, m*n))
        X = np.zeros((npts, 2))

        medxo = np.median(abs(xo))
        medyo = np.median(abs(yo))

        X[:, 0] = xi/medxo
        X[:, 1] = yi/medyo

        xx = np.zeros(m)
        xx[0] = 1.0
        for ix in range(1, m):
            xx[ix] = xx[ix-1]/medxo

        yy = np.zeros(n)
        yy[0] = 1.0
        for iy in range(1, n):
            yy[iy] = yy[iy-1]/medyo

        zz = np.outer(yy, xx)

        U = np.zeros((npts, 2))
        U[:, 0] = xo / medxo
        U[:, 1] = yo / medyo

        for iy in range(0, n):
            for ix in range(0, m):
                row = ix * n + iy
                W[:, row] = U[:, 1]**iy * U[:, 0]**ix

        # W = W.T

        WW = np.matmul(W.T, W)

        MM = linalg.inv(WW)

        MMM = np.matmul(MM.T, W.T)

        # print('MMM shape', MMM.shape)
        # print('X shape', X.shape)
        # print('zz shape', zz.shape)
        # print('n, m', n, m)

        mmx = np.matmul(MMM, X[:, 0]).reshape((m, n))
        mmy = np.matmul(MMM, X[:, 1]).reshape((m, n))

        # print('mmx shape', mmx.shape)
        # print('mmy shape', mmy.shape)

        kx = zz * mmx.T * medxo
        ky = zz * mmy.T * medyo

        # print("kx", kx)
        # print("ky", ky)

        params = np.zeros((2, plen))

        pidx = 0
        for j in range(efford):
            for ix in range(j + 1):
                iy = j - ix
                if ix <= order[1] and iy <= order[0]:
                    params[0, pidx] = kx[ix, iy]
                    params[1, pidx] = ky[ix, iy]
                else:
                    params[0, pidx] = 0.
                    params[1, pidx] = 0.
                pidx += 1

        # print("params", params)

        self.params = params

        return True

    def __call__(self, coords):
        """Apply forward transformation.

        Parameters
        ----------
        coords : (N, 2) array
            source coordinates

        Returns
        -------
        coords : (N, 2) array
            Transformed coordinates.

        """
        x = coords[:, 0]
        y = coords[:, 1]
        u = len(self.params.ravel())
        # number of coefficients -> u = (order + 1) * (order + 2)
        order = int((- 3 + math.sqrt(9 - 4 * (2 - u))) / 2)
        dst = np.zeros(coords.shape)

        pidx = 0
        for j in range(order + 1):
            for i in range(j + 1):
                dst[:, 0] += self.params[0, pidx] * x ** (j - i) * y ** i
                dst[:, 1] += self.params[1, pidx] * x ** (j - i) * y ** i
                pidx += 1

        return dst

    def inverse(self, coords):
        raise Exception(
            'There is no explicit way to do the inverse polynomial '
            'transformation. Instead, estimate the inverse transformation '
            'parameters by exchanging source and destination coordinates,'
            'then apply the forward transformation.')


TRANSFORMS = {
    'asympolynomial': AsymmetricPolynomialTransform,
}


def estimate_transform(ttype, src, dst, **kwargs):
    """Estimate 2D geometric transformation parameters.

    You can determine the over-, well- and under-determined parameters
    with the total least-squares method.

    Number of source and destination coordinates must match.

    Parameters
    ----------
    ttype : {'asympolynomial'}
        Type of transform.
    src : (N, 2) array
        Source coordinates.
    dst : (N, 2) array
        Destination coordinates.
    kwargs : array or int
        Function parameters (src, dst, n, angle)::

            NAME / TTYPE        FUNCTION PARAMETERS
            'asympolynomial'    `src, `dst`, (`order`, `order`)
                                             (x, y polynomial orders, default
                                             order is 2)

        Also see examples below.

    Returns
    -------
    tform : :class:`GeometricTransform`
        Transform object containing the transformation parameters and providing
        access to forward and inverse transformation functions.

    Examples
    --------
    >>> import numpy as np
    >>> import geometric as tf

    >>> # estimate transformation parameters
    >>> src = np.array([0, 0, 10, 10]).reshape((2, 2))
    >>> dst = np.array([12, 14, 1, -20]).reshape((2, 2))

    >>> tform = tf.estimate_transform('asympolynomial', src, dst)

    >>> np.allclose(tform.inverse(tform(src)), src)
    True

    >>> # warp image using the estimated transformation
    >>> from skimage import data
    >>> image = data.camera()

    >>> warp(image, inverse_map=tform.inverse) # doctest: +SKIP

    >>> # create transformation with explicit parameters
    >>> tform2 = tf.AsymmetriPolynomialTransform(src, dst, order=(2, 3))
    >>> np.allclose(tform2(tform(src)))
    True

    """
    ttype = ttype.lower()
    if ttype not in TRANSFORMS:
        raise ValueError('the transformation type \'%s\' is not'
                         'implemented' % ttype)

    tform = TRANSFORMS[ttype]()
    tform.estimate(src, dst, **kwargs)

    return tform



#!/usr/bin/env python
"""Modify FITS header keyword"""

from astropy.io import fits
import argparse
import sys
import os


def _parse_arguments(in_args: list) -> argparse.Namespace:

    description = "modify FITS header keyword"
    parser = argparse.ArgumentParser(prog=f"{in_args[0]}",
                                     description=description)

    parser.add_argument('infile', type=str, default=None,
                        help="Input FITS file")
    parser.add_argument('-k', '--keyword', type=str, default=None,
                        help="Keyword to modify")
    parser.add_argument('-v', '--value', type=str, default=None,
                        help="Value for keyword")
    parser.add_argument('-s', '--safe', action='store_true', default=False,
                        help="Keep original file")

    out_args = parser.parse_args(in_args[1:])
    return out_args


def main():
    args = _parse_arguments(sys.argv)

    if args.safe:
        safe_file = args.infile + '~'
        os.rename(args.infile, safe_file)
        with fits.open(safe_file) as hdul:
            hdul[0].header[args.keyword] = args.value
            hdul.writeto(args.infile)
    else:
        fits.setval(args.infile, args.keyword, value=args.value)


if __name__ == "__main__":
    main()



from keckdrpframework.primitives.base_primitive import BasePrimitive

import math


class CalcPrelimDisp(BasePrimitive):
    """
    Calculate dispersion based on configuration parameters.

    The parameters of the grating equation are calculated as:

    * BLUE

      * alpha = grating_angle - 13 - adjustment_angle
      * adjustment_angle = (180 for BH and 0 for all other gratings)

    * RED

      * alpha = 155.892 - grating_angle

    * BLUE & RED

      * beta = camera_angle - alpha

    dispersion = cos(beta)/rho/focal_length x (pixel_scale x binning) * 1.e4

    """

    def __init__(self, action, context):
        BasePrimitive.__init__(self, action, context)
        self.logger = context.pipeline_logger

    def _pre_condition(self):
        self.logger.info("Checking for master arc")
        if 'MARC' in self.action.args.ccddata.header['IMTYPE']:
            return True
        else:
            return False

    def _perform(self):
        # get binning
        y_binning = self.action.args.ybinsize
        # 0 - compute alpha
        if 'BLUE' in self.action.args.ccddata.header['CAMERA'].upper():
            preliminary_alpha = self.action.args.grangle - 13.0 - \
                self.action.args.adjang
        else:
            # red_grat_norm_angle = 156.1748047  # Caltech AIT value
            red_grat_norm_angle = 155.892
            preliminary_alpha = red_grat_norm_angle - self.action.args.grangle
        # 1 - compute preliminary angle of diffraction
        preliminary_beta = self.action.args.camangle - preliminary_alpha
        # 2 - compute preliminary dispersion
        preliminary_dispersion = math.cos(preliminary_beta/math.degrees(1.)) / \
            self.action.args.rho / self.config.instrument.FCAM * \
            (self.config.instrument.PIX*y_binning) * 1.e4
        preliminary_dispersion *= math.cos(
            self.config.instrument.GAMMA/math.degrees(1.))
        self.logger.info("Initial alpha, beta (deg): %.3f, %.3f" %
                         (preliminary_alpha, preliminary_beta))
        self.logger.info("Initial calculated dispersion (A/binned pix): %.3f" %
                         preliminary_dispersion)
        self.context.prelim_disp = preliminary_dispersion

        log_string = CalcPrelimDisp.__module__
        self.action.args.ccddata.header['HISTORY'] = log_string
        self.logger.info(log_string)

        return self.action.args
    # END: class CalcPrelimDisp()
